/*
 *   cdda - CD Digital Audio support
 *
 *   Copyright (C) 1993-2002  Ti Kan
 *   E-mail: xmcd@amb.org
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#ifndef LINT
static char *_cdda_c_ident_ = "@(#)cdda.c	7.67 02/11/01";
#endif

#include "common_d/appenv.h"
#include "common_d/util.h"
#include "libdi_d/libdi.h"
#include "cdda_d/cdda.h"
#include "cdda_d/sysvipc.h"


#ifdef USE_LIBMP3LAME
#include <stdarg.h>
#include <lame/lame.h>
#endif	/* USE_LIBMP3LAME */


extern appdata_t	app_data;
extern FILE		*errfp;


cdda_client_t		*cdda_clinfo;		/* Client information struct */

#if defined(HAS_LAME) && !defined(USE_LIBMP3LAME)
char			*lameprogs[] = {
	"lame",
	"notlame",
	NULL
};

char			*lamepath;
#endif

/*
 * Cdda method call table
 * This array must correspond to the CDDA_xxx definitions found in the
 * cdda.h file, and the index is specified by the cddaMethod parameter.
 */
struct {
	word32_t (*capab)(void);
	bool_t	 (*init)(curstat_t *);
	void	 (*halt)(di_dev_t *, curstat_t *);
	bool_t	 (*play)(di_dev_t *, curstat_t *, int, int);
	bool_t	 (*pause_resume)(di_dev_t *, curstat_t *, bool_t);
	bool_t	 (*stop)(di_dev_t *, curstat_t *);
	int	 (*vol)(di_dev_t *, curstat_t *, int, bool_t);
	bool_t	 (*chroute)(di_dev_t *, curstat_t *);
	bool_t	 (*getstatus)(di_dev_t *, curstat_t *, cdstat_t *);
	void	 (*debug)(word32_t);
	void	 (*info)(char *);
} cdda_calltbl[] = {
	{ NULL, NULL, NULL, NULL, NULL, NULL,
	  NULL, NULL, NULL, NULL, NULL },	/* CDDA_NONE */
	{ cdda_sysvipc_capab,
	  cdda_sysvipc_init,
	  cdda_sysvipc_halt,
	  cdda_sysvipc_play,
	  cdda_sysvipc_pause_resume,
	  cdda_sysvipc_stop,
	  cdda_sysvipc_vol,
	  cdda_sysvipc_chroute,
	  cdda_sysvipc_getstatus,
	  cdda_sysvipc_debug,
	  cdda_sysvipc_info }			/* CDDA_SYSVIPC */
};


/* Initialization done flag */
STATIC bool_t		cdda_initted = FALSE;

/* File format description table */
STATIC filefmt_t	cdda_ffmt_tbl[MAX_FILEFMTS] = {
	{ FILEFMT_RAW,	"RAW",	  ".raw",  "CDDA Raw (LE)"	},
	{ FILEFMT_AU,	"AU",	  ".au",   "Sun/Next Audio"	},
	{ FILEFMT_WAV,	"WAV",	  ".wav",  "Wave"		},
	{ FILEFMT_AIFF,	"AIFF",	  ".aiff", "Apple/SGI Audio"	},
	{ FILEFMT_AIFC,	"AIFF-C", ".aifc", "Apple/SGI Audio"	},
	{ FILEFMT_MP3,	"MP3",	  ".mp3",  "MPEG Layer 3"	},
	{ FILEFMT_OGG,	"OGG",	  ".ogg",  "Ogg Vorbis"		}
};

/* Internal supported bitrates table for MP3 and OggVorbis */
#define NBITRATES	16

STATIC int		_bitrates[NBITRATES] = {
	0, 32, 40, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320, -1
};

STATIC int		cdda_nbitrates = 0;


/*********************
 * Private functions *
 *********************/


/*
 * cdda_bitrates_init
 *	Initialize the number of supported bitrates for MP3 and OggVorbis
 *	output.
 *
 * Args:
 *	None.
 *
 * Return:
 *	Nothing.
 */
STATIC void
cdda_bitrates_init(void)
{
#ifdef USE_LIBMP3LAME
	int	i;

	cdda_nbitrates = 0;

	for (i = 0; i < 16 && bitrate_table[1][i] >= 0; i++)
		cdda_nbitrates++;
#else
	cdda_nbitrates = NBITRATES - 1;
#endif
}


/**********************
 *  Public functions  *
 **********************/


/*
 * cdda_capab
 *	Query configured CDDA subsystem capabilities
 *
 * Args:
 *	None.
 *
 * Return:
 *	Bitmask of supported features
 */
word32_t
cdda_capab(void)
{
	word32_t	ret,
			(*func)(void);

	if (app_data.cdda_method == CDDA_NONE)
		return 0;

	if (app_data.cdda_method < CDDA_NONE ||
	    app_data.cdda_method >= CDDA_METHODS) {
		CDDA_WARNING(app_data.str_cddainit_fail);
		DBGPRN(DBG_GEN)(errfp, "Warning: %s\n",
				  app_data.str_cddainit_fail);
		return 0;
	}

	func = cdda_calltbl[app_data.cdda_method].capab;
	if (func != NULL)
		ret = (*func)();
	else {
		CDDA_WARNING(app_data.str_cddainit_fail);
		DBGPRN(DBG_GEN)(errfp, "Warning: %s\n",
				  app_data.str_cddainit_fail);
		ret = 0;
	}

	return (ret);
}


/*
 * cdda_init
 *	Initialize the CDDA subsystem.
 *
 * Args:
 *	s   - Pointer to the curstat_t structure.
 *	clp - Pointer to cdda subsystem client registration info structure.
 *
 * Return:
 *	FALSE - failure
 *	TRUE  - success
 */
bool_t
cdda_init(curstat_t *s, cdda_client_t *clp)
{
	bool_t	ret,
		(*func)(curstat_t *);

	if (cdda_initted)
		return TRUE;

	cdda_clinfo = clp;

	if (app_data.cdda_method <= CDDA_NONE ||
	    app_data.cdda_method >= CDDA_METHODS)
		return FALSE;

	func = cdda_calltbl[app_data.cdda_method].init;
	if (func != NULL)
		ret = (*func)(s);
	else
		ret = FALSE;

	cdda_initted = ret;
	return ret;
}


/*
 * cdda_halt
 *	Shuts down the CDDA subsystem.
 *
 * Args:
 *	devp - Device descriptor
 *	s  - Pointer to the curstat_t structure
 *
 * Return:
 *	Nothing.
 */
void
cdda_halt(di_dev_t *devp, curstat_t *s)
{
	void	(*func)(di_dev_t *, curstat_t *);

	if (!cdda_initted)
		return;

	func = cdda_calltbl[app_data.cdda_method].halt;
	if (func != NULL)
		(*func)(devp, s);

	cdda_initted = FALSE;
}


/*
 * cdda_play
 *	Start playing.  No error checking is done on the output file path
 *	or pipe program strings here.  The caller should pre-check these.
 *
 * Args:
 *	devp - Device descriptor
 *	s  - Pointer to the curstat_t structure
 *	start_lba - Start logical block address
 *	end_lba   - End logical block address
 *
 * Return:
 *	FALSE - failure
 *	TRUE  - success
 */
bool_t
cdda_play(di_dev_t *devp, curstat_t *s, sword32_t start_lba, sword32_t end_lba)
{
	bool_t	(*func)(di_dev_t *, curstat_t *, sword32_t, sword32_t);

	if (!cdda_initted)
		return FALSE;

	/* If start == end, just return success */
	if (start_lba == end_lba)
		return TRUE;

	func = cdda_calltbl[app_data.cdda_method].play;
	if (func != NULL)
		return (*func)(devp, s, start_lba, end_lba);

	return FALSE;
}


/*
 * cdda_pause_resume
 *	Pause/resume playback
 *
 * Args:
 *	devp   - Device descriptor
 *	s      - Pointer to the curstat_t structure
 *	resume - Whether to resume playback
 *
 * Return:
 *	FALSE - failure
 *	TRUE  - success
 */
bool_t
cdda_pause_resume(di_dev_t *devp, curstat_t *s, bool_t resume)
{
	bool_t	(*func)(di_dev_t *, curstat_t *, bool_t);

	if (!cdda_initted)
		return FALSE;

	func = cdda_calltbl[app_data.cdda_method].pause_resume;
	if (func != NULL)
		return (*func)(devp, s, resume);

	return FALSE;
}


/*
 * cdda_stop
 *	Stop playback
 *
 * Args:
 *	devp - Device descriptor
 *	s    - Pointer to the curstat_t structure
 *
 * Return:
 *	FALSE - failure
 *	TRUE  - success
 */
bool_t
cdda_stop(di_dev_t *devp, curstat_t *s)
{
	bool_t	(*func)(di_dev_t *, curstat_t *);

	if (!cdda_initted)
		return FALSE;

	func = cdda_calltbl[app_data.cdda_method].stop;
	if (func != NULL)
		return (*func)(devp, s);

	return FALSE;
}


/*
 * cdda_vol
 *	Change volume setting
 *
 * Args:
 *	devp  - Device descriptor
 *	s     - Pointer to the curstat_t structure
 *	vol   - Desired volume level
 *	query - Whether querying or setting the volume
 *
 * Return:
 *	The volume level
 */
int
cdda_vol(di_dev_t *devp, curstat_t *s, int vol, bool_t query)
{
	int	(*func)(di_dev_t *, curstat_t *, int, bool_t);

	if (!cdda_initted)
		return FALSE;

	func = cdda_calltbl[app_data.cdda_method].vol;
	if (func != NULL)
		return (*func)(devp, s, vol, query);

	return -1;
}


/*
 * cdda_chroute
 *	Change channel routing setting
 *
 * Args:
 *	devp - Device descriptor
 *	s    - Pointer to the curstat_t structure
 *
 * Return:
 *	FALSE - failure
 *	TRUE  - success
 */
bool_t
cdda_chroute(di_dev_t *devp, curstat_t *s)
{
	bool_t	(*func)(di_dev_t *, curstat_t *);

	if (!cdda_initted)
		return FALSE;

	func = cdda_calltbl[app_data.cdda_method].chroute;
	if (func != NULL)
		return (*func)(devp, s);

	return FALSE;
}


/*
 * cdda_getstatus
 *	Get playback status
 *
 * Args:
 *	devp - Device descriptor
 *	s    - Pointer to the curstat_t structure
 *	sp - cdstat_t return info structure
 *
 * Return:
 *	FALSE - failure
 *	TRUE  - success
 */
bool_t
cdda_getstatus(di_dev_t *devp, curstat_t *s, cdstat_t *sp)
{
	bool_t	(*func)(di_dev_t *, curstat_t *, cdstat_t *);

	if (!cdda_initted)
		return FALSE;

	func = cdda_calltbl[app_data.cdda_method].getstatus;
	if (func != NULL)
		return (*func)(devp, s, sp);

	return FALSE;
}


/*
 * cdda_debug
 *	Debug level change notification function
 *
 * Args:
 *	lev - The new debug level
 *
 * Return:
 *	Nothing.
 */
void
cdda_debug(word32_t lev)
{
	void	(*func)(word32_t);

	if (!cdda_initted)
		return;

	func = cdda_calltbl[app_data.cdda_method].debug;
	if (func != NULL)
		(*func)(lev);
}


/*
 * cdda_info
 *	Obtain CDDA configuration information
 *
 * Args:
 *	None.
 *
 * Return:
 *	Pointer to an informational text string buffer.
 */
char *
cdda_info(void)
{
	void		(*func)(char *);
	static bool_t	first = TRUE;
	static char	str[STR_BUF_SZ * 2];

	if (first) {
		/* Set up info string */
		first = FALSE;
		(void) strcpy(str, "CDDA method: ");

		if (app_data.cdda_method < CDDA_NONE ||
		    app_data.cdda_method >= CDDA_METHODS) {
			(void) strcat(str, "not configured\n");
		}
		else {
			func = cdda_calltbl[app_data.cdda_method].info;
			if (func == NULL)
				(void) strcat(str, "not configured\n");
			else
				(*func)(str);
		}
	}

	return (str);
}


/*
 * cdda_filefmt_supp
 *	Given a CDDA output file format value, return a boolean indicating
 *	whether that format is supported in this compilation.
 *
 * Args:
 *	fmt - The format value
 *
 * Return:
 *	TRUE  - Format supported
 *	FALSE - Format not supported
 */
bool_t
cdda_filefmt_supp(int fmt)
{
	static bool_t	first = TRUE;

	switch (fmt) {
	case FILEFMT_RAW:
	case FILEFMT_AU:
	case FILEFMT_WAV:
	case FILEFMT_AIFF:
	case FILEFMT_AIFC:
		return TRUE;

#ifdef HAS_LAME
	case FILEFMT_MP3:
#ifdef USE_LIBMP3LAME
		return TRUE;
#else
		if (first) {
			int	i;

			first = FALSE;

			if ((lamepath = getenv("LAME_PATH")) != NULL) {
				/* Check if specified program is executable */
				if (!util_checkcmd(lamepath))
					lamepath = NULL;
			}
			else for (i = 0; lameprogs[i] != NULL; i++) {
				lamepath = util_findcmd(lameprogs[i]);
				if (lamepath != NULL)
					break;
			}

			DBGPRN(DBG_GEN|DBG_SND)(errfp,
				"\nMP3 encoder program: %s\n",
				(lamepath != NULL) ?
					lamepath :
					"not found: MP3 file format disabled"
			);
		}

		return ((bool_t) (lamepath != NULL));
#endif	/* USE_LIBMP3LAME */
#endif	/* HAS_LAME */

#ifdef HAS_VORBIS
	case FILEFMT_OGG:
		return TRUE;
#endif	/* HAS_VORBIS */

	default:
		return FALSE;
	}
	/*NOTREACHED*/
}


/*
 * cdda_filefmt
 *	Given a CDDA file format code, return a pointer to an
 *	associated filefmt_t structure.
 *
 * Args:
 *	fmt - The file format code
 *
 * Return:
 *	Pointer to the filefmt_t structure, or NULL if fmt is invalid.
 */
filefmt_t *
cdda_filefmt(int fmt)
{
	if (fmt < 0 || fmt >= MAX_FILEFMTS)
		return NULL;

	return (&cdda_ffmt_tbl[fmt]);
}


/*
 * cdda_bitrates
 *	Return the number of supported bitrates for MP3 and OggVorbis
 *
 * Args:
 *	None.
 *
 * Return:
 *	The number of supported bitrates
 */
int
cdda_bitrates(void)
{
	if (cdda_nbitrates == 0)
		cdda_bitrates_init();

	return (cdda_nbitrates);
}


/*
 * cdda_bitrate_val
 *	Given a bitrate index, return the actual bitrate value
 *
 * Args:
 *	idx - The bitrate table index
 *
 * Return:
 *	The bitrate in kb/s
 */
int
cdda_bitrate_val(int idx)
{
	if (cdda_nbitrates == 0)
		cdda_bitrates_init();

	if (idx < 0 || idx >= cdda_nbitrates)
		return -1;

#ifdef USE_LIBMP3LAME
	return (bitrate_table[1][idx]);
#else
	return (_bitrates[idx]);
#endif
}


/*
 * cdda_bitrate_name
 *	Given a bitrate index, return the actual bitrate name string
 *
 * Args:
 *	idx - The bitrate table index
 *
 * Return:
 *	Pointer to the bitrate string.  This points to an internal buffer
 *	that is overwritten on each call.
 */
char *
cdda_bitrate_name(int idx)
{
	static char	name[8];

	if (cdda_nbitrates == 0)
		cdda_bitrates_init();

	if (idx < 0 || idx >= cdda_nbitrates)
		return NULL;

#ifdef USE_LIBMP3LAME
	(void) sprintf(name, "%d", bitrate_table[1][idx]);
#else
	(void) sprintf(name, "%d", _bitrates[idx]);
#endif
	return (name);
}


/*
 * cdda_heartbeat_interval
 *	Compute the number of CDDA read/write loop iterations per
 *	hearbeat.  This is based on the read chunk size.
 *
 * Args:
 *	fps - The current transfer rate in frames per second
 *
 * Return:
 *	A integer loop iteration count.
 */
int
cdda_heartbeat_interval(int fps)
{
	return ((fps / app_data.cdda_readchkblks) + 1);
}


/*
 * cdda_heartbeat
 *	Set a heartbeat timestamp in the location specified.  This is
 *	used to check if either a CDDA read or write thread is hung or
 *	dead.
 *
 * Args:
 *	hbloc  - pointer to location where the timestamp is to be written.
 *	hbtype - CDDA_HB_READER or CDDA_HB_WRITER
 *
 * Return:
 *	Nothing.
 */
void
cdda_heartbeat(time_t *hbloc, int hbtype)
{
	if (hbloc == NULL)
		return;

	*hbloc = time(NULL);

	DBGPRN(DBG_SND)(errfp, "\nCDDA thread heartbeat: %s\n",
		(hbtype == CDDA_HB_READER) ? "reader" : "writer");
}

