/*
 *   cdda - CD Digital Audio support
 *
 *   Copyright (C) 1993-2002  Ti Kan
 *   E-mail: xmcd@amb.org
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#ifndef LINT
static char *_wr_gen_c_ident_ = "@(#)wr_gen.c	7.105 02/11/12";
#endif

#include "common_d/appenv.h"
#include "common_d/version.h"
#include "common_d/util.h"
#include "libdi_d/libdi.h"
#include "cdda_d/cdda.h"

#if defined(CDDA_SYSVIPC)

#include "cdinfo_d/cdinfo.h"
#include "cdda_d/wr_gen.h"
#include "cdda_d/au.h"
#include "cdda_d/wav.h"
#include "cdda_d/aiff.h"
#include "cdda_d/id3map.h"


#ifdef USE_LIBMP3LAME

#include <stdarg.h>
#include <lame/lame.h>		/* Use the installed lame.h */

/* LAME descriptor structure */
typedef struct {
	lame_global_flags	*lf;
	byte_t			*enc_buf;
	size_t			enc_bufsz;
} lame_desc_t;

#endif	/* USE_LIBMP3LAME */

#if defined(HAS_LAME) && !defined(USE_LIBMP3LAME)
extern char			*lamepath;	/* Path to LAME encoder */
#endif


#ifdef HAS_VORBIS

#include <vorbis/vorbisenc.h>
#include <vorbis/codec.h>

/* OggVorbis descriptor structure */
typedef struct {
	ogg_stream_state	os;
	vorbis_comment		vc;
	vorbis_info		vi;
	vorbis_dsp_state	vd;
	vorbis_block		vb;
	int			serno;
} ov_desc_t;

#endif	/* HAS_VORBIS */


/* Defines used by gen_open_pipe() */
#define STR_SHPATH	"/bin/sh"		/* Path to shell */
#define STR_SHNAME	"sh"			/* Name of shell */
#define STR_SHARG	"-c"			/* Shell arg */


extern appdata_t	app_data;
extern FILE		*errfp;
extern cdda_client_t	*cdda_clinfo;

STATIC uid_t		ruid,			/* Real user id */
			euid;			/* Effective user id */
STATIC gid_t		rgid,			/* Real group id */
			egid;			/* Effective group id */
STATIC char		*comment =
			"xmcd-" VERSION_MAJ "." VERSION_MIN "." VERSION_TEENY;
						/* File comment */


/*********************
 * Private functions *
 *********************/


/*
 * gen_esc_shellquote
 *	Handle quote characters in a quoted shell command line string.
 *
 * Args:
 *	gdp - Pointer to the gen_desc_t structure
 *	str - Input string
 *
 * Return:
 *	Return string.  This is a dynamically allocated string.  The caller
 *	should deallocate it via MEM_FREE when done.  A NULL is returned
 *	if the input string is NULL, or if the output string buffer cannot
 *	be allocated.
 */
STATIC char *
gen_esc_shellquote(gen_desc_t *gdp, char *str)
{
	char	*ret,
		*cp;

	if (str == NULL)
		return NULL;

	if ((ret = (char *) MEM_ALLOC("ret", strlen(str) + 1)) == NULL) {
		(void) sprintf(gdp->cdp->i->msgbuf,
			    "gen_esc_shellquote: Out of memory.");
		DBGPRN(DBG_GEN)(errfp, "%s\n", gdp->cdp->i->msgbuf);
		return NULL;
	}

	for (cp = ret; *str != '\0'; cp++, str++) {
		if (*str == '\047')
			*cp = '`';	/* Change single quote to backquote */
		else
			*cp = *str;
	}
	*cp = '\0';

	return (ret);
}


/*
 * gen_seek
 *	Wrapper function for fseek(3) and lseek(2).  It uses the appropriate
 *	function based on whether a buffered stdio stream is opened.
 *
 * Args:
 *	gdp    - Pointer to the gen_desc_t structure
 *	offset - Seek offset
 *	whence - Seek type (SEEK_SET, SEEK_CUR or SEEK_END)
 *
 * Return:
 *	FALSE - failure
 *	TRUE  - success
 */
STATIC bool_t
gen_seek(gen_desc_t *gdp, off_t offset, int whence)
{
	int	ret;

	if (gdp->fp != NULL) {
		if ((ret = fseek(gdp->fp, offset, whence)) < 0) {
			(void) sprintf(gdp->cdp->i->msgbuf,
					"gen_seek: %s: fseek failed",
					gdp->path);
			DBGPRN(DBG_SND)(errfp, "%s\n", gdp->cdp->i->msgbuf);
		}
	}
	else {
		if ((ret = lseek(gdp->fd, offset, whence)) < 0) {
			(void) sprintf(gdp->cdp->i->msgbuf,
					"gen_seek: %s: lseek failed",
					gdp->path);
			DBGPRN(DBG_SND)(errfp, "%s\n", gdp->cdp->i->msgbuf);
		}
	}

	return ((bool_t) (ret >= 0));
}


/*
 * gen_genre_cddb2id3
 *	Map CDDB2 genre to ID3tag genre.  This function now does a sparse
 *	mapping because the two genre systems are not one-to-one mappable.
 *	Moreover this has to work for both classic CDDB and CDDB2.
 * 
 * Args:
 *	genreid - CDDB genre ID string
 *
 * Return:
 *	ID3tag genre ID string.
 */
STATIC char *
gen_genre_cddb2id3(char	*genreid)
{
	cdinfo_genre_t	*gp;
	char		*id;
	int		i;

	if ((gp = cdinfo_genre(genreid)) == NULL)
		return ("12");	/* "Other" */

	if (gp->parent == NULL)
		id = gp->id;		/* CDDB Genre */
	else
		id = gp->parent->id;	/* CDDB Subgenre */

	/* Look for mapping in the genre mapping table */
	for (i = 0; id3_gmap[i].cddbgenre != NULL; i++) {
		if (strcmp(id, id3_gmap[i].cddbgenre) == 0)
			return (id3_gmap[i].id3genre);
	}

	return ("12");	/* Other */
}


/*
 * gen_write_au_hdr
 *	Writes an au file header to the output file.
 *
 * Args:
 *	gdp - Pointer to the gen_desc_t structure
 *
 * Return:
 *	FALSE - failure
 *	TRUE  - success
 */
STATIC bool_t
gen_write_au_hdr(gen_desc_t *gdp)
{
	au_filehdr_t	hdr;

	if (gdp->datalen == 0) {
		struct stat	stbuf;

		/* Get file information */
		if (fstat(gdp->fd, &stbuf) < 0) {
			(void) sprintf(gdp->cdp->i->msgbuf,
				"gen_write_au_hdr: fstat failed (errno=%d)",
				errno);
			DBGPRN(DBG_SND)(errfp, "%s\n", gdp->cdp->i->msgbuf);
			return FALSE;
		}

		/* Set actual data length */
		gdp->datalen = stbuf.st_size - sizeof(hdr) - strlen(comment);

		/* Rewind to beginning of file */
		if (!gen_seek(gdp, (off_t) 0, SEEK_SET))
			return FALSE;
	}

	(void) memset(&hdr, 0, sizeof(hdr));

	/* Initialize an au file header. */
	hdr.au_magic = util_bswap32(AUDIO_AU_FILE_MAGIC);
	hdr.au_offset = util_bswap32(sizeof(hdr) + strlen(comment));
	hdr.au_data_size = util_bswap32((word32_t) gdp->datalen);
	hdr.au_encoding = util_bswap32(AUDIO_AU_ENCODING_LINEAR_16);
	hdr.au_sample_rate = util_bswap32(44100);
	hdr.au_channels = util_bswap32(2);

	/* Write header */
	if (!gen_write_chunk(gdp, (byte_t *) &hdr, sizeof(hdr)))
		return FALSE;

	/* Write comment */
	if (!gen_write_chunk(gdp, (byte_t *) comment, strlen(comment)))
		return FALSE;

	return TRUE;
}


/*
 * gen_write_wav_hdr
 *	Writes an wav file header to the output file.
 *
 * Args:
 *	gdp - Pointer to the gen_desc_t structure
 *
 * Return:
 *	FALSE - failure
 *	TRUE  - success
 */
STATIC bool_t
gen_write_wav_hdr(gen_desc_t *gdp)
{
	wav_filehdr_t	hdr;

	if (gdp->datalen == 0) {
		struct stat	stbuf;

		/* Get file information */
		if (fstat(gdp->fd, &stbuf) < 0) {
			(void) sprintf(gdp->cdp->i->msgbuf,
				"gen_write_wav_hdr: fstat failed (errno=%d)",
				errno);
			DBGPRN(DBG_SND)(errfp, "%s\n", gdp->cdp->i->msgbuf);
			return FALSE;
		}

		/* Set actual data length */
		gdp->datalen = stbuf.st_size - sizeof(hdr);

		/* Rewind to beginning of file */
		if (!gen_seek(gdp, (off_t) 0, SEEK_SET))
			return FALSE;
	}

	(void) memset(&hdr, 0, sizeof(hdr));

	/* RIFF chunk */
	(void) strncpy(hdr.r_riff, "RIFF", 4);
	hdr.r_length = util_lswap32((word32_t) gdp->datalen + 36);
	(void) strncpy(hdr.r_wave, "WAVE", 4);

	/* FORMAT chunk */
	(void) strncpy(hdr.f_format, "fmt ", 4);
	hdr.f_length = util_lswap32(16);
	hdr.f_const = util_lswap16(1);
	hdr.f_channels = util_lswap16(2);
	hdr.f_sample_rate = util_lswap32(44100);
	hdr.f_bytes_per_s = util_lswap32(44100 * 2 * 2);
	hdr.f_bytes_per_sample = util_lswap16(2);
	hdr.f_bits_per_sample = util_lswap16(16);

	/* DATA chunk */
	(void) strncpy(hdr.d_data, "data", 4);
	hdr.d_length = util_lswap32((word32_t) gdp->datalen);

	/* Write header */
	if (!gen_write_chunk(gdp, (byte_t *) &hdr, sizeof(hdr)))
		return FALSE;

	return TRUE;
}


/*
 * gen_write_aiff_hdr
 *	Writes an aiff file header to the output file.
 *
 * Args:
 *	gdp - Pointer to the gen_desc_t structure
 *
 * Return:
 *	FALSE - failure
 *	TRUE  - success
 */
STATIC bool_t
gen_write_aiff_hdr(gen_desc_t *gdp)
{
	word32_t	tmp;
	aiff_filehdr_t	hdr;

	if (gdp->datalen == 0) {
		struct stat	stbuf;

		/* Get file information */
		if (fstat(gdp->fd, &stbuf) < 0) {
			(void) sprintf(gdp->cdp->i->msgbuf,
				"gen_write_aiff_hdr: fstat failed (errno=%d)",
				errno);
			DBGPRN(DBG_SND)(errfp, "%s\n", gdp->cdp->i->msgbuf);
			return FALSE;
		}

		/* Set actual data length */
		gdp->datalen = stbuf.st_size - AIFF_HDRSZ;

		/* Rewind to beginning of file */
		if (!gen_seek(gdp, (off_t) 0, SEEK_SET))
			return FALSE;
	}

	(void) memset(&hdr, 0, sizeof(hdr));

	/* FORM chunk */
	(void) strncpy(hdr.a_form, "FORM", 4);
	tmp = (word32_t) gdp->datalen + AIFF_HDRSZ - 8;
	hdr.a_length[0] = (tmp & 0xff000000) >> 24;
	hdr.a_length[1] = (tmp & 0x00ff0000) >> 16;
	hdr.a_length[2] = (tmp & 0x0000ff00) >> 8;
	hdr.a_length[3] = (tmp & 0x000000ff);
	(void) strncpy(hdr.a_aiff, "AIFF", 4);

	/* COMM chunk */
	(void) strncpy(hdr.c_comm, "COMM", 4);
	tmp = 18;
	hdr.c_length[0] = (tmp & 0xff000000) >> 24;
	hdr.c_length[1] = (tmp & 0x00ff0000) >> 16;
	hdr.c_length[2] = (tmp & 0x0000ff00) >> 8;
	hdr.c_length[3] = (tmp & 0x000000ff);
	tmp = 2;
	hdr.c_channels[0] = (tmp & 0xff00) >> 8;
	hdr.c_channels[1] = (tmp & 0x00ff);
	tmp = (word32_t) gdp->datalen >> 2;
	hdr.c_frames[0] = (tmp & 0xff000000) >> 24;
	hdr.c_frames[1] = (tmp & 0x00ff0000) >> 16;
	hdr.c_frames[2] = (tmp & 0x0000ff00) >> 8;
	hdr.c_frames[3] = (tmp & 0x000000ff);
	tmp = 16;
	hdr.c_sample_size[0] = (tmp & 0xff00) >> 8;
	hdr.c_sample_size[1] = (tmp & 0x00ff);
	(void) strncpy(hdr.c_sample_rate, "@\016\254D\0\0\0\0\0\0", 10);
				/* 44100 in 80-bit IEEE floating point */

	/* SSND chunk */
	(void) strncpy(hdr.s_ssnd, "SSND", 4);
	tmp = (word32_t) gdp->datalen + 8;
	hdr.s_length[0] = (tmp & 0xff000000) >> 24;
	hdr.s_length[1] = (tmp & 0x00ff0000) >> 16;
	hdr.s_length[2] = (tmp & 0x0000ff00) >> 8;
	hdr.s_length[3] = (tmp & 0x000000ff);

	/* Write header */
	if (!gen_write_chunk(gdp, (byte_t *) &hdr, AIFF_HDRSZ))
		return FALSE;

	return TRUE;
}


/*
 * gen_write_aifc_hdr
 *	Writes an aifc file header to the output file.
 *
 * Args:
 *	gdp - Pointer to the gen_desc_t structure
 *
 * Return:
 *	FALSE - failure
 *	TRUE  - success
 */
STATIC bool_t
gen_write_aifc_hdr(gen_desc_t *gdp)
{
	word32_t	tmp;
	aifc_filehdr_t	hdr;

	if (gdp->datalen == 0) {
		struct stat	stbuf;

		/* Get file information */
		if (fstat(gdp->fd, &stbuf) < 0) {
			(void) sprintf(gdp->cdp->i->msgbuf,
				"gen_write_aifc_hdr: fstat failed (errno=%d)",
				errno);
			DBGPRN(DBG_SND)(errfp, "%s\n", gdp->cdp->i->msgbuf);
			return FALSE;
		}

		/* Set actual data length */
		gdp->datalen = stbuf.st_size - AIFC_HDRSZ;

		/* Rewind to beginning of file */
		if (!gen_seek(gdp, (off_t) 0, SEEK_SET))
			return FALSE;
	}

	(void) memset(&hdr, 0, sizeof(hdr));

	/* FORM chunk */
	(void) strncpy(hdr.a_form, "FORM", 4);
	tmp = (word32_t) gdp->datalen + AIFC_HDRSZ - 8;
	hdr.a_length[0] = (tmp & 0xff000000) >> 24;
	hdr.a_length[1] = (tmp & 0x00ff0000) >> 16;
	hdr.a_length[2] = (tmp & 0x0000ff00) >> 8;
	hdr.a_length[3] = (tmp & 0x000000ff);
	(void) strncpy(hdr.a_aifc, "AIFC", 4);

	/* FVER chunk */
	(void) strncpy(hdr.f_fver, "FVER", 4);
	tmp = 4;
	hdr.f_length[0] = (tmp & 0xff000000) >> 24;
	hdr.f_length[1] = (tmp & 0x00ff0000) >> 16;
	hdr.f_length[2] = (tmp & 0x0000ff00) >> 8;
	hdr.f_length[3] = (tmp & 0x000000ff);
	tmp = 2726318400UL;
	hdr.f_version[0] = (tmp & 0xff000000) >> 24;
	hdr.f_version[1] = (tmp & 0x00ff0000) >> 16;
	hdr.f_version[2] = (tmp & 0x0000ff00) >> 8;
	hdr.f_version[3] = (tmp & 0x000000ff);

	/* COMM chunk */
	(void) strncpy(hdr.c_comm, "COMM", 4);
	tmp = 18;
	hdr.c_length[0] = (tmp & 0xff000000) >> 24;
	hdr.c_length[1] = (tmp & 0x00ff0000) >> 16;
	hdr.c_length[2] = (tmp & 0x0000ff00) >> 8;
	hdr.c_length[3] = (tmp & 0x000000ff);
	tmp = 2;
	hdr.c_channels[0] = (tmp & 0xff00) >> 8;
	hdr.c_channels[1] = (tmp & 0x00ff);
	tmp = (word32_t) gdp->datalen >> 2;
	hdr.c_frames[0] = (tmp & 0xff000000) >> 24;
	hdr.c_frames[1] = (tmp & 0x00ff0000) >> 16;
	hdr.c_frames[2] = (tmp & 0x0000ff00) >> 8;
	hdr.c_frames[3] = (tmp & 0x000000ff);
	tmp = 16;
	hdr.c_sample_size[0] = (tmp & 0xff00) >> 8;
	hdr.c_sample_size[1] = (tmp & 0x00ff);
	(void) strncpy(hdr.c_sample_rate, "@\016\254D\0\0\0\0\0\0", 10);
				/* 44100 in 80-bit IEEE floating point */
	(void) strncpy(hdr.c_comptype, "NONE", 4);
	hdr.c_complength = 14;
	(void) strncpy(hdr.c_compstr, "not compressed", 14);

	/* SSND chunk */
	(void) strncpy(hdr.s_ssnd, "SSND", 4);
	tmp = (word32_t) gdp->datalen + 8;
	hdr.s_length[0] = (tmp & 0xff000000) >> 24;
	hdr.s_length[1] = (tmp & 0x00ff0000) >> 16;
	hdr.s_length[2] = (tmp & 0x0000ff00) >> 8;
	hdr.s_length[3] = (tmp & 0x000000ff);

	/* Write header */
	if (!gen_write_chunk(gdp, (byte_t *) &hdr, AIFC_HDRSZ))
		return FALSE;

	return TRUE;
}


#ifdef USE_LIBMP3LAME
/*
 * gen_add_id3tags
 *	Initialize and fill the MP3 ID3 tags information with data from CDDB.
 *
 * Args:
 *	gdp - Pointer to the gen_desc_t structure
 *
 * Return:
 *	Nothing.
 */
STATIC void
gen_add_id3tags(gen_desc_t *gdp)
{
	lame_desc_t	*ldesc = (lame_desc_t *) gdp->aux;
	curstat_t	*s = cdda_clinfo->curstat_addr();
	cdinfo_incore_t	*cdp = cdinfo_addr();
	int		idx = (int) gdp->cdp->i->trk_idx;
	char		trkstr[4];

	if (ldesc == NULL)
		return;

	if ((gdp->flags & GDESC_ISPIPE) && app_data.id3tag_mode == ID3TAG_V2)
		return;

	id3tag_init(ldesc->lf);

	if (app_data.id3tag_mode == ID3TAG_V1) {
		id3tag_v1_only(ldesc->lf);
	}
	else if ((gdp->flags & GDESC_ISPIPE) == 0) {
		if (app_data.id3tag_mode == ID3TAG_V2)
			id3tag_v2_only(ldesc->lf);
		else
			id3tag_add_v2(ldesc->lf);
	}

	DBGPRN(DBG_SND)(errfp, "\nSetting ID3 tags for %s\n", gdp->path);

	if (cdp->disc.title != NULL)
		id3tag_set_album(ldesc->lf, cdp->disc.title);

	if (app_data.cdda_trkfile && cdp->track[idx].artist != NULL)
		id3tag_set_artist(ldesc->lf, cdp->track[idx].artist);
	else if (cdp->disc.artist != NULL)
		id3tag_set_artist(ldesc->lf, cdp->disc.artist);

	if (app_data.cdda_trkfile && cdp->track[idx].title != NULL)
		id3tag_set_title(ldesc->lf, cdp->track[idx].title);
	else if (cdp->disc.title != NULL)
		id3tag_set_title(ldesc->lf, cdp->disc.title);

	if (app_data.cdda_trkfile && cdp->track[idx].year != NULL)
		id3tag_set_year(ldesc->lf, cdp->track[idx].year);
	else if (cdp->disc.year != NULL)
		id3tag_set_year(ldesc->lf, cdp->disc.year);

	if (app_data.cdda_trkfile && cdp->track[idx].genre != NULL)
		(void) id3tag_set_genre(ldesc->lf,
			gen_genre_cddb2id3(cdp->track[idx].genre)
		);
	else if (cdp->disc.genre != NULL)
		(void) id3tag_set_genre(ldesc->lf,
			gen_genre_cddb2id3(cdp->disc.genre)
		);

	if (app_data.cdda_trkfile) {
		(void) sprintf(trkstr, "%d", (int) s->trkinfo[idx].trkno);
		id3tag_set_track(ldesc->lf, trkstr);
	}

	(void) id3tag_set_comment(ldesc->lf, comment);
}
#endif	/* USE_LIBMP3LAME */


#ifdef HAS_VORBIS
/*
 * gen_add_vtags
 *	Initialize and fill the Vorbis comment information with data
 *	from CDDB.
 *
 * Args:
 *	gdp - Pointer to the gen_desc_t structure
 *
 * Return:
 *	Nothing.
 */
STATIC void
gen_add_vtags(gen_desc_t *gdp)
{
	ov_desc_t	*ovdesc = (ov_desc_t *) gdp->aux;
	curstat_t	*s = cdda_clinfo->curstat_addr();
	cdinfo_incore_t	*cdp = cdinfo_addr();
	int		idx = (int) gdp->cdp->i->trk_idx;
	char		trkstr[4];

	if (ovdesc == NULL)
		return;

	DBGPRN(DBG_SND)(errfp, "\nSetting comment tags for %s\n", gdp->path);

	vorbis_comment_init(&ovdesc->vc);

	if (cdp->disc.title != NULL)
		vorbis_comment_add_tag(&ovdesc->vc, "album", cdp->disc.title);

	if (app_data.cdda_trkfile && cdp->track[idx].artist != NULL) {
		vorbis_comment_add_tag(
			&ovdesc->vc, "artist",
			cdp->track[idx].artist
		);
	}
	else if (cdp->disc.artist != NULL) {
		vorbis_comment_add_tag(
			&ovdesc->vc, "artist",
			cdp->disc.artist
		);
	}

	if (app_data.cdda_trkfile && cdp->track[idx].title != NULL) {
		vorbis_comment_add_tag(
			&ovdesc->vc, "title",
			cdp->track[idx].title
		);
	}
	else if (cdp->disc.title != NULL) {
		vorbis_comment_add_tag(
			&ovdesc->vc, "title",
			cdp->disc.title
		);
	}

	if (app_data.cdda_trkfile && cdp->track[idx].year != NULL) {
		vorbis_comment_add_tag(
			&ovdesc->vc, "date",
			cdp->track[idx].year
		);
	}
	else if (cdp->disc.year != NULL) {
		vorbis_comment_add_tag(
			&ovdesc->vc, "date",
			cdp->disc.year
		);
	}

	if (app_data.cdda_trkfile && cdp->track[idx].genre != NULL) {
		vorbis_comment_add_tag(
			&ovdesc->vc, "genre",
			cdinfo_genre_name(cdp->track[idx].genre)
		);
	}
	else if (cdp->disc.genre != NULL) {
		vorbis_comment_add_tag(
			&ovdesc->vc, "genre",
			cdinfo_genre_name(cdp->disc.genre)
		);
	}

	if (app_data.cdda_trkfile) {
		(void) sprintf(trkstr, "%d", (int) s->trkinfo[idx].trkno);
		vorbis_comment_add_tag(&ovdesc->vc, "tracknumber", trkstr);
	}

	vorbis_comment_add_tag(&ovdesc->vc, "Encoder", comment);
}
#endif	/* HAS_VORBIS */


#if defined(HAS_LAME) && !defined(USE_LIBMP3LAME)
/*
 * gen_lame_cmd
 *	Create a command string to invoke the LAME encoder program.
 *
 * Args:
 *	gdp - Pointer to the gen_desc_t structure.
 *
 * Return:
 *	The command string.  This is a dynamically allocated buffer.  The
 *	caller should use MEM_FREE to deallocate it when done.  If there
 *	is insufficient memory, NULL is returned.
 */
STATIC char *
gen_lame_cmd(gen_desc_t *gdp)
{
	char		*cmd,
			*cp;
	curstat_t	*s = cdda_clinfo->curstat_addr();

	cmd = (char *) MEM_ALLOC("cmd",
		strlen(lamepath) + strlen(gdp->path) + strlen(comment) + 400
	);
	if (cmd == NULL)
		return NULL;

	(void) sprintf(cmd, "%s%s -r -s 44.1 --bitwidth 16",
		lamepath,
		((app_data.debug & (DBG_GEN|DBG_SND)) == 0) ? " -S" : ""
	);

	if (gdp->fmt == FILEFMT_OGG)
		(void) strcat(cmd, " --ogginput");

	switch (app_data.comp_mode) {
	case COMPMODE_CBR:
		(void) sprintf(cmd, "%s --cbr -b %d", cmd, app_data.bitrate);
		break;
	case COMPMODE_VBR:
		(void) sprintf(cmd, "%s --vbr-old -V %d", cmd,
				-(app_data.qual_factor - 10));
		if (app_data.bitrate_min > 0)
			(void) sprintf(cmd, "%s -b %d", cmd,
					app_data.bitrate_min);
		if (app_data.bitrate_max > 0)
			(void) sprintf(cmd, "%s -B %d", cmd,
					app_data.bitrate_max);
		break;
	case COMPMODE_VBR2:
		(void) sprintf(cmd, "%s --vbr-new -V %d", cmd,
				-(app_data.qual_factor - 10));
		if (app_data.bitrate_min > 0)
			(void) sprintf(cmd, "%s -b %d", cmd,
					app_data.bitrate_min);
		if (app_data.bitrate_max > 0)
			(void) sprintf(cmd, "%s -B %d", cmd,
					app_data.bitrate_max);
		break;
	case COMPMODE_ABR:
		(void) sprintf(cmd, "%s --abr %d", cmd, app_data.bitrate);
		if (app_data.bitrate_min > 0)
			(void) sprintf(cmd, "%s -b %d", cmd,
					app_data.bitrate_min);
		if (app_data.bitrate_max > 0)
			(void) sprintf(cmd, "%s -B %d", cmd,
					app_data.bitrate_max);
		break;
	default:
		DBGPRN(DBG_GEN)(errfp,
			"NOTICE: Invalid compressionMode parameter.\n");
		break;
	}

	(void) sprintf(cmd, "%s -q %d", cmd, -(app_data.mp3_algo - 10));

	switch (app_data.mp3_mode) {
	case MP3_STEREO:
		(void) strcat(cmd, " -m s");
		break;
	case MP3_JSTEREO:
		(void) strcat(cmd, " -m j");
		break;
	case MP3_FORCEMS:
		(void) strcat(cmd, " -m f");
		break;
	case MP3_MONO:
		(void) strcat(cmd, " -m s -a");
		break;
	default:
		DBGPRN(DBG_GEN)(errfp,
			"NOTICE: Invalid mp3Mode parameter: Set to 0.\n");
		(void) strcat(cmd, " -m s");
		break;
	}

	if (app_data.lowpass_mode == FILTER_OFF &&
	    app_data.highpass_mode == FILTER_OFF)
		(void) strcat(cmd, " -k");
	else {
		if (app_data.lowpass_mode == FILTER_MANUAL) {
			(void) sprintf(cmd,
				"%s --lowpass %.1f --lowpass-width %.1f",
				cmd,
				(float) app_data.lowpass_freq / 1024.0,
				(float) app_data.lowpass_width / 1024.0
			);
		}
		if (app_data.highpass_mode == FILTER_MANUAL) {
			(void) sprintf(cmd,
				"%s --highpass %.1f --highpass-width %.1f",
				cmd,
				(float) app_data.highpass_freq / 1024.0,
				(float) app_data.highpass_width / 1024.0
			);
		}
	}

	if (app_data.copyright)
		(void) strcat(cmd, " -c");
	if (!app_data.original)
		(void) strcat(cmd, " -o");
	if (app_data.checksum)
		(void) strcat(cmd, " -p");
	if (app_data.nores)
		(void) strcat(cmd, " --nores");
	if (app_data.strict_iso)
		(void) strcat(cmd, " --strictly-enforce-ISO");

	if (app_data.add_tag) {
		cdinfo_incore_t	*cdp = cdinfo_addr();
		int		idx = (int) gdp->cdp->i->trk_idx;

		switch (app_data.id3tag_mode) {
		case ID3TAG_V1:
			(void) strcat(cmd, " --id3v1-only");
			break;
		case ID3TAG_V2:
			(void) strcat(cmd, " --id3v2-only");
			break;
		case ID3TAG_BOTH:
			(void) strcat(cmd, " --add-id3v2");
			break;
		default:
			DBGPRN(DBG_GEN)(errfp,
				"NOTICE: Invalid id3TagMode parameter.\n");
			break;
		}

		if (cdp->disc.title != NULL) {
			cp = gen_esc_shellquote(gdp, cdp->disc.title);
			if (cp == NULL)
				return NULL;
			(void) sprintf(cmd, "%s --tl '%.80s'", cmd, cp);
			MEM_FREE(cp);
		}

		if (app_data.cdda_trkfile && cdp->track[idx].artist != NULL) {
			cp = gen_esc_shellquote(gdp, cdp->track[idx].artist);
			if (cp == NULL)
				return NULL;
			(void) sprintf(cmd, "%s --ta '%.80s'", cmd, cp);
			MEM_FREE(cp);
		}
		else if (cdp->disc.artist != NULL) {
			cp = gen_esc_shellquote(gdp, cdp->disc.artist);
			if (cp == NULL)
				return NULL;
			(void) sprintf(cmd, "%s --ta '%.80s'", cmd, cp);
			MEM_FREE(cp);
		}

		if (app_data.cdda_trkfile && cdp->track[idx].title != NULL) {
			cp = gen_esc_shellquote(gdp, cdp->track[idx].title);
			if (cp == NULL)
				return NULL;
			(void) sprintf(cmd, "%s --tt '%.80s'", cmd, cp);
			MEM_FREE(cp);
		}
		else if (cdp->disc.title != NULL) {
			cp = gen_esc_shellquote(gdp, cdp->disc.title);
			if (cp == NULL)
				return NULL;
			(void) sprintf(cmd, "%s --tt '%.80s'", cmd, cp);
			MEM_FREE(cp);
		}

		if (app_data.cdda_trkfile && cdp->track[idx].year != NULL) {
			(void) sprintf(cmd, "%s --ty '%.4s'", cmd,
					cdp->track[idx].year);
		}
		else if (cdp->disc.year != NULL) {
			(void) sprintf(cmd, "%s --ty '%.4s'", cmd,
					cdp->disc.year);
		}

		if (app_data.cdda_trkfile && cdp->track[idx].genre != NULL) {
			(void) sprintf(cmd, "%s --tg '%.4s'", cmd,
				    gen_genre_cddb2id3(cdp->track[idx].genre));
		}
		else if (cdp->disc.genre != NULL) {
			(void) sprintf(cmd, "%s --tg '%.4s'", cmd,
				    gen_genre_cddb2id3(cdp->disc.genre));
		}

		if (app_data.cdda_trkfile) {
			(void) sprintf(cmd, "%s --tn '%d'", cmd,
					(int) s->trkinfo[idx].trkno);
		}
		(void) sprintf(cmd, "%s --tc '%.40s'", cmd, comment);
	}

	if ((gdp->flags & GDESC_ISPIPE) != 0) {
		/* gdp->path is the program to pipe to */
		(void) sprintf(cmd,
			"%s - - 2>%s | %s",
			cmd,
			((app_data.debug & (DBG_SND|DBG_GEN)) != 0) ?
				"/tmp/xmcd-lame.$$" : "/dev/null",
			gdp->path
		);
	}
	else {
		/* gdp->path is the file to write to */
		if ((cp = gen_esc_shellquote(gdp, gdp->path)) == NULL)
			return NULL;

		(void) sprintf(cmd,
			"%s - '%s' >%s 2>&1 && chmod %o '%s'",
			cmd,
			cp,
			((app_data.debug & (DBG_SND|DBG_GEN)) != 0) ?
				"/tmp/xmcd-lame.$$" : "/dev/null",
			(unsigned int) gdp->mode,
			cp
		);

		MEM_FREE(cp);
	}

	return (cmd);
}
#endif	/* HAS_LAME !USE_LIBMP3LAME */


#ifdef USE_LIBMP3LAME
/*
 * gen_init_lame
 *	Initialize LAME encoder and set up encoding parameters
 *
 * Args:
 *	gdp - Pointer to the gen_desc_t structure
 *
 * Return:
 *	FALSE - failure
 *	TRUE  - success
 */
/*ARGSUSED*/
STATIC bool_t
gen_init_lame(gen_desc_t *gdp)
{
	int		n,
			qual,
			algo = -(app_data.mp3_algo - 10);
	lame_desc_t	*ldesc;

	/* Allocate a LAME descriptor structure */
	ldesc = (lame_desc_t *) MEM_ALLOC("lame_desc_t", sizeof(lame_desc_t));
	if (ldesc == NULL) {
		(void) sprintf(gdp->cdp->i->msgbuf,
			"gen_init_lame: Out of memory.");
		DBGPRN(DBG_GEN)(errfp, "%s\n", gdp->cdp->i->msgbuf);
		return FALSE;
	}
	(void) memset(ldesc, 0, sizeof(lame_desc_t));

	gdp->aux = (void *) ldesc;

	/* Init LAME decoder */
	if ((ldesc->lf = lame_init()) == NULL) {
		(void) sprintf(gdp->cdp->i->msgbuf,
			"gen_init_lame: Encoder initialization failed.");
		DBGPRN(DBG_SND)(errfp, "%s\n", gdp->cdp->i->msgbuf);
		return FALSE;
	}

	/* Use library built-in message functions which prints to stderr */
	(void) lame_set_msgf(ldesc->lf, NULL);
	(void) lame_set_debugf(ldesc->lf, NULL);
	(void) lame_set_errorf(ldesc->lf, NULL);

	/* Input sampling rate and number of channels */
	(void) lame_set_in_samplerate(ldesc->lf, 44100);
	(void) lame_set_num_channels(ldesc->lf, 2);

	/* Encoding method and parameters */
	switch (app_data.comp_mode) {
	case COMPMODE_CBR:
		(void) lame_set_VBR(ldesc->lf, vbr_off);
		if (app_data.bitrate > 0)
			(void) lame_set_brate(ldesc->lf, app_data.bitrate);
		break;

	case COMPMODE_VBR:
		qual = -(app_data.qual_factor - 10);

		(void) lame_set_VBR(ldesc->lf, vbr_rh);
		(void) lame_set_VBR_q(ldesc->lf, qual);
		if (app_data.bitrate_min > 0) {
			(void) lame_set_brate(ldesc->lf, app_data.bitrate_min);
			(void) lame_set_VBR_min_bitrate_kbps(
				ldesc->lf, app_data.bitrate_min
			);
		}
		if (app_data.bitrate_max > 0) {
			(void) lame_set_VBR_max_bitrate_kbps(
				ldesc->lf, app_data.bitrate_max
			);
		}
		break;

	case COMPMODE_VBR2:
		qual = -(app_data.qual_factor - 10);

		/* algo greater than 7 has no psymodel,
		 * and VBR2 needs it, so limit it to 7
		 */
		if (algo > 7)
			algo = 7;

		(void) lame_set_VBR(ldesc->lf, vbr_mtrh);
		(void) lame_set_VBR_q(ldesc->lf, qual);
		if (app_data.bitrate_min > 0) {
			(void) lame_set_brate(ldesc->lf, app_data.bitrate_min);
			(void) lame_set_VBR_min_bitrate_kbps(
				ldesc->lf, app_data.bitrate_min
			);
		}
		if (app_data.bitrate_max > 0) {
			(void) lame_set_VBR_max_bitrate_kbps(
				ldesc->lf, app_data.bitrate_max
			);
		}
		break;

	case COMPMODE_ABR:
		(void) lame_set_VBR(ldesc->lf, vbr_abr);
		if (app_data.bitrate > 0) {
			(void) lame_set_VBR_mean_bitrate_kbps(
				ldesc->lf, app_data.bitrate
			);
		}
		if (app_data.bitrate_min > 0) {
			(void) lame_set_brate(ldesc->lf, app_data.bitrate_min);
			(void) lame_set_VBR_min_bitrate_kbps(
				ldesc->lf, app_data.bitrate_min
			);
		}
		if (app_data.bitrate_max > 0) {
			(void) lame_set_VBR_max_bitrate_kbps(
				ldesc->lf, app_data.bitrate_max
			);
		}
		break;

	default:
		DBGPRN(DBG_GEN)(errfp,
			"NOTICE: Invalid compressionMode parameter.\n");
		break;
	}

	/* Noise shaping and psychoacoustics */
	(void) lame_set_quality(ldesc->lf, algo);

	/* Stereo/mono modes */
	switch (app_data.mp3_mode) {
	case MP3_STEREO:
		(void) lame_set_mode(ldesc->lf, STEREO);
		break;
	case MP3_FORCEMS:
		(void) lame_set_force_ms(ldesc->lf, 1);
		/*FALLTHROUGH*/
	case MP3_JSTEREO:
		(void) lame_set_mode(ldesc->lf, JOINT_STEREO);
		break;
	case MP3_MONO:
		(void) lame_set_mode(ldesc->lf, MONO);
		break;
	default:
		DBGPRN(DBG_GEN)(errfp,
			"NOTICE: Invalid mp3Mode parameter: Set to 0.\n");
		(void) lame_set_mode(ldesc->lf, STEREO);
		break;
	}

	/* Set lowpass filter if specified */
	switch (app_data.lowpass_mode) {
	case FILTER_OFF:
		(void) lame_set_lowpassfreq(ldesc->lf, -1);
		break;
	case FILTER_AUTO:
		(void) lame_set_lowpassfreq(ldesc->lf, 0);
		break;
	case FILTER_MANUAL:
		(void) lame_set_lowpassfreq(
			ldesc->lf, app_data.lowpass_freq
		);
		(void) lame_set_lowpasswidth(
			ldesc->lf, app_data.lowpass_width
		);
		break;
	default:
		DBGPRN(DBG_GEN)(errfp,
		    "NOTICE: Invalid lowpassMode parameter: Set to 0.\n");
		(void) lame_set_lowpassfreq(ldesc->lf, -1);
		break;
	}

	/* Set highpass filter if specified */
	switch (app_data.highpass_mode) {
	case FILTER_OFF:
		(void) lame_set_highpassfreq(ldesc->lf, -1);
		break;
	case FILTER_AUTO:
		(void) lame_set_highpassfreq(ldesc->lf, 0);
		break;
	case FILTER_MANUAL:
		(void) lame_set_highpassfreq(
			ldesc->lf, app_data.highpass_freq
		);
		(void) lame_set_highpasswidth(
			ldesc->lf, app_data.highpass_width
		);
		break;
	default:
		DBGPRN(DBG_GEN)(errfp,
		    "NOTICE: Invalid highpassMode parameter: Set to 0.\n");
		(void) lame_set_highpassfreq(ldesc->lf, -1);
		break;
	}

	/* Set flags */
	(void) lame_set_copyright(ldesc->lf, (int) app_data.copyright);
	(void) lame_set_original(ldesc->lf, (int) app_data.original);
	if (app_data.nores) {
		(void) lame_set_disable_reservoir(ldesc->lf, 1);
		(void) lame_set_padding_type(ldesc->lf, PAD_NO);
	}
	(void) lame_set_error_protection(ldesc->lf, (int) app_data.checksum);
	(void) lame_set_strict_ISO(ldesc->lf, (int) app_data.strict_iso);

	/* If not a pipe, set up to write a VBR tag */
	(void) lame_set_bWriteVbrTag(
		ldesc->lf, ((gdp->flags & GDESC_ISPIPE) == 0) ? 1 : 0
	);

	if (gdp->fmt == FILEFMT_MP3 && app_data.add_tag)
		gen_add_id3tags(gdp);

	if (lame_init_params(ldesc->lf) < 0) {
		(void) lame_close(ldesc->lf);
		(void) sprintf(gdp->cdp->i->msgbuf,
			    "gen_init_lame: Encoder parameters init failed.");
		DBGPRN(DBG_SND)(errfp, "%s\n", gdp->cdp->i->msgbuf);
		return FALSE;
	}

	/* Allocate memory for the encoder output buffer */
	n = app_data.cdda_readchkblks * (44100 / 75);
	ldesc->enc_bufsz = (size_t) (1.25 * n) + 7200;

	ldesc->enc_buf = (byte_t *) MEM_ALLOC("enc_buf", ldesc->enc_bufsz);
	if (ldesc->enc_buf == NULL) {
		(void) lame_close(ldesc->lf);
		(void) sprintf(gdp->cdp->i->msgbuf,
			    "gen_init_lame: Out of memory.");
		DBGPRN(DBG_GEN)(errfp, "%s\n", gdp->cdp->i->msgbuf);
		return FALSE;
	}

	return TRUE;
}


/*
 * gen_encode_lame_chunk
 *	Encodes the audio data using LAME and write to the fd.
 *
 * Args:
 *	gdp  - Pointer to the gen_desc_t structure
 *	data - Pointer to data
 *	len  - Length of data in bytes
 *
 * Return:
 *	FALSE - failure
 *	TRUE  - success
 */
/*ARGSUSED*/
STATIC bool_t
gen_encode_lame_chunk(gen_desc_t *gdp, byte_t *data, size_t len)
{
	lame_desc_t	*ldesc = (lame_desc_t *) gdp->aux;
	int		samples,
			n;

	if (ldesc == NULL || ldesc->lf == NULL)
		return FALSE;

	samples = len >> 2;

	DBGPRN(DBG_SND)(errfp, "\nEncoding %d samples\n", samples);

	n = lame_encode_buffer_interleaved(
		ldesc->lf, (sword16_t *)(void *) data, samples,
		ldesc->enc_buf, (int) ldesc->enc_bufsz
	);

	if (n < 0) {
		(void) sprintf(gdp->cdp->i->msgbuf,
				"gen_encode_lame_chunk: "
				"Encode failed (status=%d)",
				n);
		DBGPRN(DBG_SND)(errfp, "%s\n", gdp->cdp->i->msgbuf);
		return FALSE;
	}

	/* Write out the output buffer */
	gdp->flags |= GDESC_WRITEOUT;
	return (gen_write_chunk(gdp, ldesc->enc_buf, n));
}


/*
 * gen_halt_lame
 *	Flush buffers and shut down LAME encoder
 *
 * Args:
 *	gdp - Pointer to the gen_desc_t structure
 *
 * Return:
 *	FALSE - failure
 *	TRUE  - success
 */
/*ARGSUSED*/
STATIC bool_t
gen_halt_lame(gen_desc_t *gdp)
{
	lame_desc_t	*ldesc = (lame_desc_t *) gdp->aux;
	int	n;

	if (ldesc == NULL)
		return FALSE;

	n = lame_encode_flush(ldesc->lf, ldesc->enc_buf, ldesc->enc_bufsz);
	if (n > 0) {
		gdp->flags |= GDESC_WRITEOUT;
		(void) gen_write_chunk(gdp, ldesc->enc_buf, n);
	}

	/* If not a pipe, add a Xing VBR tag */
	if ((gdp->flags & GDESC_ISPIPE) == 0)
		lame_mp3_tags_fid(ldesc->lf, gdp->fp);

	(void) lame_close(ldesc->lf);

	MEM_FREE(ldesc->enc_buf);
	MEM_FREE(ldesc);
	gdp->aux = NULL;

	return TRUE;
}
#endif	/* USE_LIBMP3LAME */


#ifdef HAS_VORBIS
/*
 * gen_init_vorbis
 *	Initialize OggVorbis encoder and set up encoding parameters
 *
 * Args:
 *	gdp - Pointer to the gen_desc_t structure
 *
 * Return:
 *	FALSE - failure
 *	TRUE  - success
 */
/*ARGSUSED*/
STATIC bool_t
gen_init_vorbis(gen_desc_t *gdp)
{
	int		bitrate,
			minbr,
			maxbr;
	float		qual;
	ov_desc_t	*ovdesc;
	ogg_page	og;
	ogg_packet	hdr_main;
	ogg_packet	hdr_comments;
	ogg_packet	hdr_codebooks;
	static bool_t	first = TRUE;

	/* Allocate a OggVorbis descriptor structure */
	ovdesc = (ov_desc_t *) MEM_ALLOC("ov_desc_t", sizeof(ov_desc_t));
	if (ovdesc == NULL) {
		(void) sprintf(gdp->cdp->i->msgbuf,
			"gen_init_vorbis: Out of memory.");
		DBGPRN(DBG_GEN)(errfp, "%s\n", gdp->cdp->i->msgbuf);
		return FALSE;
	}
	(void) memset(ovdesc, 0, sizeof(ov_desc_t));

	gdp->aux = (void *) ovdesc;

	if (first) {
		first = FALSE;
		srand(time(NULL));
		ovdesc->serno = rand();
	}

	if (app_data.add_tag)
		gen_add_vtags(gdp);

	vorbis_info_init(&ovdesc->vi);

	switch (app_data.comp_mode) {
	case COMPMODE_CBR:
	case COMPMODE_ABR:
		bitrate = (app_data.bitrate > 0) ?
			  (app_data.bitrate * 1000) : 128000;
		minbr = (app_data.bitrate_min > 0) ?
			(app_data.bitrate_min * 1000) : -1;
		maxbr = (app_data.bitrate_max > 0) ?
			(app_data.bitrate_max * 1000) : -1;

		/* Managed bitrate */
		if (vorbis_encode_init(&ovdesc->vi, 2, 44100,
				       minbr, bitrate, maxbr) < 0) {
			vorbis_info_clear(&ovdesc->vi);
			(void) sprintf(gdp->cdp->i->msgbuf,
					"gen_init_vorbis: "
					"Encoder init (CBR/ABR) failed.");
			DBGPRN(DBG_SND)(errfp, "%s\n", gdp->cdp->i->msgbuf);
			return FALSE;
		}
		break;

	case COMPMODE_VBR:
	case COMPMODE_VBR2:
		qual = (float) app_data.qual_factor / 10.0f;

		/* Variable bitrate */
		if (vorbis_encode_init_vbr(&ovdesc->vi, 2, 44100, qual) < 0) {
			vorbis_info_clear(&ovdesc->vi);
			(void) sprintf(gdp->cdp->i->msgbuf,
					"gen_init_vorbis: "
					"Encoder init (VBR) failed.");
			DBGPRN(DBG_SND)(errfp, "%s\n", gdp->cdp->i->msgbuf);
			return FALSE;
		}
		break;

	default:
		DBGPRN(DBG_GEN)(errfp,
			"NOTICE: Invalid compressionMode parameter.\n");
		break;
	}

	vorbis_analysis_init(&ovdesc->vd, &ovdesc->vi);
	vorbis_block_init(&ovdesc->vd, &ovdesc->vb);
	ogg_stream_init(&ovdesc->os, ovdesc->serno++);

	/* Build the header packets and stream them out */
	vorbis_analysis_headerout(
		&ovdesc->vd,
		&ovdesc->vc,
		&hdr_main,
		&hdr_comments,
		&hdr_codebooks
	);
	ogg_stream_packetin(&ovdesc->os, &hdr_main);
	ogg_stream_packetin(&ovdesc->os, &hdr_comments);
	ogg_stream_packetin(&ovdesc->os, &hdr_codebooks);

	while (ogg_stream_flush(&ovdesc->os, &og) != 0) {
		gdp->flags |= GDESC_WRITEOUT;
		if (!gen_write_chunk(gdp, (byte_t *) og.header, og.header_len))
			return FALSE;

		gdp->flags |= GDESC_WRITEOUT;
		if (!gen_write_chunk(gdp, (byte_t *) og.body, og.body_len))
			return FALSE;
	}

	return TRUE;
}


/*
 * gen_encode_vorbis_chunk
 *	Encodes the audio data using OggVorbis and write to the fd.
 *
 * Args:
 *	gdp  - Pointer to the gen_desc_t structure
 *	data - Pointer to data
 *	len  - Data length in bytes
 *
 * Return:
 *	FALSE - failure
 *	TRUE  - success
 */
/*ARGSUSED*/
STATIC bool_t
gen_encode_vorbis_chunk(gen_desc_t *gdp, byte_t *data, size_t len)
{
	ov_desc_t	*ovdesc = (ov_desc_t *) gdp->aux;
	ogg_page	og;
	ogg_packet	op;
	int		samples,
			i,
			j,
			ret;
	sbyte_t		*p;
	bool_t		eos;

	if (ovdesc == NULL)
		return FALSE;

	samples = len >> 2;

	DBGPRN(DBG_SND)(errfp, "\nEncoding %d samples\n", samples);

	if (samples == 0 || data == NULL) {
		vorbis_analysis_wrote(&ovdesc->vd, 0);
	}
	else {
		float	**venc_buf;

		venc_buf = vorbis_analysis_buffer(&ovdesc->vd, samples);

		/* Un-interleave audio data into encode buffer,
		 * assume little endian.
		 */
		p = (sbyte_t *) data;
		for (i = 0; i < samples; i++) {
		    j = i << 2;
		    venc_buf[0][i]=
			((p[j+1] << 8) | (0x00ff & (int) p[j])) / 32768.0f;
		    venc_buf[1][i]=
			((p[j+3] << 8) | (0x00ff & (int) p[j+2])) / 32768.0f;
		}

		vorbis_analysis_wrote(&ovdesc->vd, samples);
	}

	eos = FALSE;
	while (vorbis_analysis_blockout(&ovdesc->vd, &ovdesc->vb) == 1) {
		/* Do the main analysis, creating a packet */
		vorbis_analysis(&ovdesc->vb, NULL);
		vorbis_bitrate_addblock(&ovdesc->vb);

		while (vorbis_bitrate_flushpacket(&ovdesc->vd, &op)) {
			/* Add packet to bitstream */
			ogg_stream_packetin(&ovdesc->os, &op);

			while (!eos) {
				ret = ogg_stream_pageout(&ovdesc->os, &og);
				if (ret == 0)
					break;

				gdp->flags |= GDESC_WRITEOUT;
				if (!gen_write_chunk(gdp, (byte_t *) og.header,
						     og.header_len))
					return FALSE;

				gdp->flags |= GDESC_WRITEOUT;
				if (!gen_write_chunk(gdp, (byte_t *) og.body,
						     og.body_len))
					return FALSE;

				if (ogg_page_eos(&og))
					eos = TRUE;
			}
		}
	}

	return TRUE;
}


/*
 * gen_halt_vorbis
 *	Flush buffers and shut down OggVorbis encoder
 *
 * Args:
 *	gdp - Pointer to the gen_desc_t structure
 *
 * Return:
 *	FALSE - failure
 *	TRUE  - success
 */
/*ARGSUSED*/
STATIC bool_t
gen_halt_vorbis(gen_desc_t *gdp)
{
	ov_desc_t	*ovdesc = (ov_desc_t *) gdp->aux;

	if (ovdesc == NULL)
		return FALSE;

	/* Signal end of output */
	(void) gen_encode_vorbis_chunk(gdp, NULL, 0);

	ogg_stream_clear(&ovdesc->os);
	vorbis_block_clear(&ovdesc->vb);
	vorbis_dsp_clear(&ovdesc->vd);
	vorbis_comment_clear(&ovdesc->vc);
	vorbis_info_clear(&ovdesc->vi);

	MEM_FREE(ovdesc);
	gdp->aux = NULL;

	return TRUE;
}
#endif	/* HAS_VORBIS */


/********************
 * Public functions *
 ********************/


/*
 * gen_set_eid
 *	Set effective uid/gid to that of the invoking user.  This is
 *	called before accessing any device or files for security reasons.
 *
 * Args:
 *	cdp - Pointer to the cd_state_t structure
 *
 * Return:
 *	FALSE - failure
 *	TRUE  - success
 */
bool_t
gen_set_eid(cd_state_t *cdp)
{
#ifdef HAS_SETEUID
	if (seteuid(ruid) < 0 || geteuid() != ruid) {
		(void) sprintf(cdp->i->msgbuf, "gen_set_eid: Cannot set uid");
		DBGPRN(DBG_GEN)(errfp, "%s\n", cdp->i->msgbuf);
		return FALSE;
	}
	if (setegid(rgid) < 0 || getegid() != rgid) {
		(void) sprintf(cdp->i->msgbuf, "gen_set_eid: Cannot set gid");
		DBGPRN(DBG_GEN)(errfp, "%s\n", cdp->i->msgbuf);
		return FALSE;
	}
#endif
	return TRUE;
}


/*
 * gen_reset_eid
 *	Restore saved uid/gid after the use of gen_set_eid()..
 *
 * Args:
 *	cdp - Pointer to the cd_state_t structure
 *
 * Return:
 *	FALSE - failure
 *	TRUE  - success
 */
bool_t
gen_reset_eid(cd_state_t *cdp)
{
#ifdef HAS_SETEUID
	if (seteuid(euid) < 0 || geteuid() != euid) {
		(void) sprintf(cdp->i->msgbuf,
				"gen_reset_eid: Cannot set uid");
		DBGPRN(DBG_GEN)(errfp, "%s\n", cdp->i->msgbuf);
		return FALSE;
	}
	if (setegid(egid) < 0 || getegid() != egid) {
		(void) sprintf(cdp->i->msgbuf,
				"gen_reset_eid: Cannot set gid");
		DBGPRN(DBG_GEN)(errfp, "%s\n", cdp->i->msgbuf);
		return FALSE;
	}
#endif

	return TRUE;
}


/*
 * gen_open_file
 *	Open an output file.
 *
 * Args:
 *	cdp	- Pointer to the cd_state_t structure
 *	path	- File path name
 *	oflag	- Open flags
 *	mode	- Open mode
 *	fmt	- File header format to be written
 *	datalen	- Audio data length in bytes
 *
 * Return:
 *	Pointer to a gen_desc_t structure associated with this open,
 *	or NULL on failure.
 */
gen_desc_t *
gen_open_file(
	cd_state_t	*cdp,
	char		*path,
	int		oflag,
	mode_t		mode,
	int		fmt,
	size_t		datalen
)
{
	gen_desc_t	*gdp;
	int		fd;
	char		*oflag_str;
	struct stat	stbuf;

	if (!cdda_filefmt_supp(fmt)) {
		(void) sprintf(cdp->i->msgbuf,
			    "gen_open_file: Support for the selected file "
			    "format (%d) is not compiled-in.",
			    fmt);
		DBGPRN(DBG_SND)(errfp, "%s\n", cdp->i->msgbuf);
		return NULL;
	}

	if (!gen_set_eid(cdp))
		return NULL;

	if ((oflag & 0x03) == O_WRONLY && (oflag & O_CREAT) != 0) {
		char	*dpath;
		mode_t	dmode;

		if ((dpath = util_dirname(path)) == NULL) {
			(void) sprintf(cdp->i->msgbuf,
				"gen_open_file: Invalid directory path:\n%s",
				dpath);
			DBGPRN(DBG_SND)(errfp, "%s\n", cdp->i->msgbuf);
			(void) gen_reset_eid(cdp);
			return NULL;
		}

		if (stat(dpath, &stbuf) < 0 ) {
			if (errno == ENOENT) {
				/* Set directory perm based on file perm */
				dmode = (mode | S_IXUSR);
				if (mode & S_IRGRP)
					dmode |= (S_IRGRP | S_IXGRP);
				if (mode & S_IWGRP)
					dmode |= (S_IWGRP | S_IXGRP);
				if (mode & S_IROTH)
					dmode |= (S_IROTH | S_IXOTH);
				if (mode & S_IWOTH)
					dmode |= (S_IWOTH | S_IXOTH);

				/* Make directory if necessary */
				if (!util_mkdir(dpath, dmode)) {
					(void) sprintf(cdp->i->msgbuf,
						"gen_open_file: "
						"Cannot create directory:\n%s",
						dpath
					);
					DBGPRN(DBG_SND)(errfp, "%s\n",
							cdp->i->msgbuf);
					MEM_FREE(dpath);
					(void) gen_reset_eid(cdp);
					return NULL;
				}
			}
			else {
				(void) sprintf(cdp->i->msgbuf,
					"gen_open_file: "
					"Cannot stat %s (errno=%d)",
					dpath,
					errno
				);
				DBGPRN(DBG_SND)(errfp, "%s\n", cdp->i->msgbuf);
				MEM_FREE(dpath);
				(void) gen_reset_eid(cdp);
				return NULL;
			}
		}
		else if (!S_ISDIR(stbuf.st_mode)) {
			(void) sprintf(cdp->i->msgbuf,
				    "gen_open_file: %s:\nNot a directory.",
				    dpath);
			DBGPRN(DBG_SND)(errfp, "%s\n", cdp->i->msgbuf);
			MEM_FREE(dpath);
			(void) gen_reset_eid(cdp);
			return NULL;
		}

		MEM_FREE(dpath);
	}

#if defined(HAS_LAME) && !defined(USE_LIBMP3LAME)
	switch (fmt) {
	case FILEFMT_MP3:
		{
			char	*cmd;

			/* Use the external lame encoder, pipe audio
			 * data to it.
			 */
			if (lamepath == NULL) {
				(void) sprintf(cdp->i->msgbuf,
					    "LAME encoder program not found.");
				DBGPRN(DBG_GEN)(errfp, "%s\n", cdp->i->msgbuf);
				(void) gen_reset_eid(cdp);
				return NULL;
			}

			/* Create temporary gen_desc_t structure */
			gdp = (gen_desc_t *) MEM_ALLOC(
				"gen_desc_t", sizeof(gen_desc_t)
			);
			if (gdp == NULL) {
				(void) sprintf(cdp->i->msgbuf,
					    "gen_open_file: Out of memory.");
				DBGPRN(DBG_GEN)(errfp, "%s\n", cdp->i->msgbuf);
				(void) gen_reset_eid(cdp);
				return NULL;
			}
			(void) memset(gdp, 0, sizeof(gen_desc_t));

			gdp->fd = -1;
			gdp->mode = mode;
			gdp->fmt = fmt;
			gdp->cdp = cdp;
			gdp->datalen = datalen;
			gdp->fp = NULL;

			if (!util_newstr(&gdp->path, path)) {
				MEM_FREE(gdp);
				(void) sprintf(cdp->i->msgbuf,
					    "gen_open_file: Out of memory.");
				DBGPRN(DBG_GEN)(errfp, "%s\n", cdp->i->msgbuf);
				(void) gen_reset_eid(cdp);
				return NULL;
			}

			if ((cmd = gen_lame_cmd(gdp)) == NULL) {
				MEM_FREE(gdp->path);
				MEM_FREE(gdp);
				(void) sprintf(cdp->i->msgbuf,
					    "gen_open_file: Out of memory.");
				DBGPRN(DBG_GEN)(errfp, "%s\n", cdp->i->msgbuf);
				(void) gen_reset_eid(cdp);
				return NULL;
			}

			MEM_FREE(gdp->path);
			MEM_FREE(gdp);

			/* Spawn lame encoder and open pipe to it */
			gdp = gen_open_pipe(cdp, cmd, NULL, fmt, datalen);

			MEM_FREE(cmd);

			(void) gen_reset_eid(cdp);
			return (gdp);
		}
		/*NOTREACHED*/

	default:
		break;
	}
#endif	/* HAS_LAME !USE_LIBMP3LAME */

	/* Open file */
	if ((fd = open(path, oflag, mode)) < 0) {
		(void) sprintf(cdp->i->msgbuf,
			    "gen_open_file: open of %s failed (errno=%d)",
			    path, errno);
		DBGPRN(DBG_SND)(errfp, "%s\n", cdp->i->msgbuf);
		(void) gen_reset_eid(cdp);
		return NULL;
	}

	DBGPRN(DBG_GEN|DBG_SND)(errfp, "\nOpened [%s]\n", path);

	gdp = (gen_desc_t *) MEM_ALLOC("gen_desc_t", sizeof(gen_desc_t));
	if (gdp == NULL) {
		(void) close(fd);
		(void) sprintf(cdp->i->msgbuf,
			    "gen_open_file: Out of memory.");
		DBGPRN(DBG_GEN)(errfp, "%s\n", cdp->i->msgbuf);
		(void) gen_reset_eid(cdp);
		return NULL;
	}
	(void) memset(gdp, 0, sizeof(gen_desc_t));

	switch ((int) (oflag & 0x03)) {
	case O_WRONLY:
		oflag_str = "w";
		(void) chmod(path, mode);
		break;
	case O_RDWR:
		oflag_str = "r+";
		(void) chmod(path, mode);
		break;
	case O_RDONLY:
	default:
		oflag_str = "r";
		break;
	}

	gdp->fd = fd;
	gdp->mode = mode;
	gdp->fmt = fmt;
	gdp->cdp = cdp;
	gdp->datalen = datalen;

	if (fstat(gdp->fd, &stbuf) == 0 &&
	    (S_ISCHR(stbuf.st_mode) || S_ISBLK(stbuf.st_mode))) {
		gdp->fp = NULL;	/* Use non-stdio on devices special files */
	}
	else {
		if ((gdp->fp = fdopen(gdp->fd, oflag_str)) == NULL) {
			DBGPRN(DBG_SND)(errfp,
					"%s: fdopen failed.\n"
					"Using non-buffered file I/O.\n",
					path);
		}
		else {
			gdp->buf = (char *) MEM_ALLOC("gdp_buf", GDESC_BUFSZ);
			if (gdp->buf != NULL)
				setbuf(gdp->fp, gdp->buf);
		}
	}

	if (!util_newstr(&gdp->path, path)) {
		if (gdp->fp != NULL)
			(void) fclose(gdp->fp);
		else
			(void) close(gdp->fd);
		MEM_FREE(gdp);
		(void) sprintf(cdp->i->msgbuf,
			    "gen_open_file: Out of memory.");
		DBGPRN(DBG_GEN)(errfp, "%s\n", cdp->i->msgbuf);
		(void) gen_reset_eid(cdp);
		return NULL;
	}

	/* Update file header */
	switch (fmt) {
	case FILEFMT_AU:
		if (!gen_write_au_hdr(gdp)) {
			(void) close(fd);
			MEM_FREE(gdp->path);
			MEM_FREE(gdp);
			(void) gen_reset_eid(cdp);
			return NULL;
		}
		break;

	case FILEFMT_WAV:
		if (!gen_write_wav_hdr(gdp)) {
			(void) close(fd);
			MEM_FREE(gdp->path);
			MEM_FREE(gdp);
			(void) gen_reset_eid(cdp);
			return NULL;
		}
		break;

	case FILEFMT_AIFF:
		if (!gen_write_aiff_hdr(gdp)) {
			(void) close(fd);
			MEM_FREE(gdp->path);
			MEM_FREE(gdp);
			(void) gen_reset_eid(cdp);
			return NULL;
		}
		break;

	case FILEFMT_AIFC:
		if (!gen_write_aifc_hdr(gdp)) {
			(void) close(fd);
			MEM_FREE(gdp->path);
			MEM_FREE(gdp);
			(void) gen_reset_eid(cdp);
			return NULL;
		}
		break;

#ifdef USE_LIBMP3LAME
	case FILEFMT_MP3:
		if (!gen_init_lame(gdp)) {
			(void) close(fd);
			MEM_FREE(gdp->path);
			MEM_FREE(gdp);
			(void) gen_reset_eid(cdp);
			return NULL;
		}
		break;
#endif	/* USE_LIBMP3LAME */

#ifdef HAS_VORBIS
	case FILEFMT_OGG:
		if (!gen_init_vorbis(gdp)) {
			(void) close(fd);
			MEM_FREE(gdp->path);
			MEM_FREE(gdp);
			(void) gen_reset_eid(cdp);
			return NULL;
		}
		break;
#endif	/* HAS_VORBIS */

	default:
		break;
	}

	(void) gen_reset_eid(cdp);
	return (gdp);
}


/*
 * gen_close_file
 *	Close a file opened with gen_open_file().
 *
 * Args:
 *	gdp - Pointer to the gen_desc_t structure
 *
 * Return:
 *	FALSE - failure
 *	TRUE  - success
 */
bool_t
gen_close_file(gen_desc_t *gdp)
{
	cd_state_t	*cdp = gdp->cdp;
	bool_t		ret;

	if (!gen_set_eid(cdp))
		return FALSE;

	gdp->datalen = 0;

	switch (gdp->fmt) {
	case FILEFMT_AU:
		(void) gen_write_au_hdr(gdp);
		break;

	case FILEFMT_WAV:
		(void) gen_write_wav_hdr(gdp);
		break;

	case FILEFMT_AIFF:
		(void) gen_write_aiff_hdr(gdp);
		break;

	case FILEFMT_AIFC:
		(void) gen_write_aifc_hdr(gdp);
		break;

#ifdef HAS_LAME
	case FILEFMT_MP3:
#ifdef USE_LIBMP3LAME
		(void) gen_halt_lame(gdp);
		break;
#else
		ret = gen_close_pipe(gdp);
		(void) gen_reset_eid(cdp);
		return (ret);
#endif	/* USE_LIBMP3LAME */
#endif	/* HAS_LAME */

#ifdef HAS_VORBIS
	case FILEFMT_OGG:
		(void) gen_halt_vorbis(gdp);
		break;
#endif	/* HAS_VORBIS */

	default:
		break;
	}

	/* Close file */
	if (gdp->fp != NULL)
		ret = (fclose(gdp->fp) == 0);
	else
		ret = (close(gdp->fd) == 0);

	if (ret) {
		DBGPRN(DBG_GEN|DBG_SND)(errfp, "\nClosed [%s]\n", gdp->path);
	}
	else {
		DBGPRN(DBG_GEN|DBG_SND)(errfp,
			"\nClose [%s] failed: errno=%d\n",
			gdp->path, errno);
	}

	MEM_FREE(gdp->path);
	MEM_FREE(gdp);

	(void) gen_reset_eid(cdp);
	return (ret);
}


/*
 * gen_open_pipe
 *	Spawn an external program and connect a pipe to its stdin
 *
 * Args:
 *	cdp	 - Pointer to the cd_state_t structure
 *	progpath - Command line of program to spawn
 *	ret_pid	 - Child process ID return
 *	fmt	 - File header format to be written
 *	datalen	 - Audio data length in bytes
 *
 * Return:
 *	Pointer to a gen_desc_t structure associated with this open,
 *	or NULL on failure.
 */
gen_desc_t *
gen_open_pipe(
	cd_state_t	*cdp,
	char		*progpath,
	pid_t		*ret_pid,
	int		fmt,
	size_t		datalen
)
{
	gen_desc_t	*gdp;
	char		*cmd;
	pid_t		cpid;
	int		i,
			fd[2];

	if (!cdda_filefmt_supp(fmt)) {
		(void) sprintf(cdp->i->msgbuf,
			    "gen_open_pipe: Support for the selected file "
			    "format (%d) is not compiled-in.",
			    fmt);
		DBGPRN(DBG_SND)(errfp, "%s\n", cdp->i->msgbuf);
		return NULL;
	}

	if (!gen_set_eid(cdp))
		return NULL;

	if (PIPE(fd) < 0) {
		(void) sprintf(cdp->i->msgbuf,
				"gen_open_pipe: output pipe failed (errno=%d)",
				errno);
		DBGPRN(DBG_SND)(errfp, "%s\n", cdp->i->msgbuf);
		(void) gen_reset_eid(cdp);
		return NULL;
	}

	cmd = NULL;
	if (!util_newstr(&cmd, progpath)) {
		(void) sprintf(cdp->i->msgbuf,
			    "gen_open_pipe: Out of memory.");
		DBGPRN(DBG_GEN)(errfp, "%s\n", cdp->i->msgbuf);
		(void) gen_reset_eid(cdp);
		return NULL;
	}

#if defined(HAS_LAME) && !defined(USE_LIBMP3LAME)
	if (ret_pid != NULL) {
		switch (fmt) {
		case FILEFMT_MP3:
			/* Use the external lame encoder, pipe audio
			 * data to it.
			 */
			if (lamepath == NULL) {
				(void) sprintf(cdp->i->msgbuf,
					    "LAME encoder program not found.");
				DBGPRN(DBG_GEN)(errfp, "%s\n", cdp->i->msgbuf);
				(void) gen_reset_eid(cdp);
				return NULL;
			}

			/* Create temporary gen_desc_t structure */
			gdp = (gen_desc_t *) MEM_ALLOC(
				"gen_desc_t", sizeof(gen_desc_t)
			);
			if (gdp == NULL) {
				(void) sprintf(cdp->i->msgbuf,
					    "gen_open_file: Out of memory.");
				DBGPRN(DBG_GEN)(errfp, "%s\n", cdp->i->msgbuf);
				(void) gen_reset_eid(cdp);
				return NULL;
			}
			(void) memset(gdp, 0, sizeof(gen_desc_t));

			gdp->fd = -1;
			gdp->fmt = fmt;
			gdp->cdp = cdp;
			gdp->datalen = datalen;
			gdp->fp = NULL;
			gdp->flags |= GDESC_ISPIPE;

			if (!util_newstr(&gdp->path, progpath)) {
				MEM_FREE(gdp);
				(void) sprintf(cdp->i->msgbuf,
					    "gen_open_pipe: Out of memory.");
				DBGPRN(DBG_GEN)(errfp, "%s\n", cdp->i->msgbuf);
				(void) gen_reset_eid(cdp);
				return NULL;
			}

			if (cmd != NULL)
				MEM_FREE(cmd);

			if ((cmd = gen_lame_cmd(gdp)) == NULL) {
				MEM_FREE(gdp->path);
				MEM_FREE(gdp);
				(void) sprintf(cdp->i->msgbuf,
					    "gen_open_pipe: Out of memory.");
				DBGPRN(DBG_GEN)(errfp, "%s\n", cdp->i->msgbuf);
				(void) gen_reset_eid(cdp);
				return NULL;
			}

			MEM_FREE(gdp->path);
			MEM_FREE(gdp);
			break;

		default:
			break;
		}
	}
#endif	/* HAS_LAME !USE_LIBMP3LAME */

	DBGPRN(DBG_GEN|DBG_SND)(errfp, "\nInvoking [%s]\n", cmd);

	switch (cpid = FORK()) {
	case -1:
		(void) sprintf(cdp->i->msgbuf,
			    "gen_open_pipe: fork failed (errno=%d)",
			    errno);
		DBGPRN(DBG_GEN)(errfp, "%s\n", cdp->i->msgbuf);
		(void) gen_reset_eid(cdp);
		return NULL;

	case 0:
		/* Child */

		/* Close unneeded pipe descriptors */
		(void) close(fd[1]);

		(void) close(0);	/* Close stdin */

		for (i = 3; i < 255; i++) {
			/* Close unneeded fds */
			if (i != fd[0])
				(void) close(i);
		}

		if (dup(fd[0]) < 0) {	/* Connect stdin to pipe */
			(void) fprintf(errfp,
				    "gen_open_pipe: dup failed (errno=%d)\n",
				    errno);
			exit(1);
		}

		(void) signal(SIGTERM, SIG_DFL);

		/* Exec the pipe program */
		(void) execl(STR_SHPATH, STR_SHNAME, STR_SHARG, cmd, NULL);

		/* An error has occurred */
		exit(errno);
		/*NOTREACHED*/

	default:
		/* Parent */
		if (ret_pid != NULL)
			*ret_pid = cpid;

		/* Close unneeded pipe descriptor */
		(void) close(fd[0]);
		break;
	}

	gdp = (gen_desc_t *) MEM_ALLOC("gen_desc_t", sizeof(gen_desc_t));
	if (gdp == NULL) {
		(void) close(fd[1]);
		(void) sprintf(cdp->i->msgbuf,
			    "gen_open_pipe: Out of memory.");
		DBGPRN(DBG_GEN)(errfp, "%s\n", cdp->i->msgbuf);
		(void) gen_reset_eid(cdp);
		return NULL;
	}
	(void) memset(gdp, 0, sizeof(gen_desc_t));

	gdp->fd = fd[1];
	gdp->fmt = fmt;
	gdp->cdp = cdp;
	gdp->datalen = datalen;
	gdp->flags |= GDESC_ISPIPE;
	gdp->path = cmd;
	gdp->cpid = cpid;

	if ((gdp->fp = fdopen(gdp->fd, "w")) == NULL) {
		DBGPRN(DBG_SND)(errfp,
				"%s: fdopen failed.\n"
				"Using non-buffered file I/O.\n",
				gdp->path);
	}
	else {
		gdp->buf = (char *) MEM_ALLOC("gdp_buf", GDESC_BUFSZ);
		if (gdp->buf != NULL)
			setbuf(gdp->fp, gdp->buf);
	}

	switch (fmt) {
	case FILEFMT_AU:
		if (!gen_write_au_hdr(gdp)) {
			(void) close(fd[1]);
			MEM_FREE(gdp->path);
			MEM_FREE(gdp);
			(void) gen_reset_eid(cdp);
			return NULL;
		}
		break;

	case FILEFMT_WAV:
		if (!gen_write_wav_hdr(gdp)) {
			(void) close(fd[1]);
			MEM_FREE(gdp->path);
			MEM_FREE(gdp);
			(void) gen_reset_eid(cdp);
			return NULL;
		}
		break;

	case FILEFMT_AIFF:
		if (!gen_write_aiff_hdr(gdp)) {
			(void) close(fd[1]);
			MEM_FREE(gdp->path);
			MEM_FREE(gdp);
			(void) gen_reset_eid(cdp);
			return NULL;
		}
		break;

	case FILEFMT_AIFC:
		if (!gen_write_aifc_hdr(gdp)) {
			(void) close(fd[1]);
			MEM_FREE(gdp->path);
			MEM_FREE(gdp);
			(void) gen_reset_eid(cdp);
			return NULL;
		}
		break;

#ifdef USE_LIBMP3LAME
	case FILEFMT_MP3:
		if (!gen_init_lame(gdp)) {
			(void) close(fd[1]);
			MEM_FREE(gdp->path);
			MEM_FREE(gdp);
			(void) gen_reset_eid(cdp);
			return NULL;
		}
		break;
#endif	/* USE_LIBMP3LAME */

#ifdef HAS_VORBIS
	case FILEFMT_OGG:
		if (!gen_init_vorbis(gdp)) {
			(void) close(fd[1]);
			MEM_FREE(gdp->path);
			MEM_FREE(gdp);
			(void) gen_reset_eid(cdp);
			return NULL;
		}
		break;
#endif	/* HAS_VORBIS */

	default:
		break;
	}

	(void) gen_reset_eid(cdp);
	return (gdp);
}


/*
 * gen_close_pipe
 *	Close the program pipe that was opened via gen_open_pipe()
 *
 * Args:
 *	gdp - Pointer to the gen_desc_t structure
 *
 * Return:
 *	FALSE - failure
 *	TRUE  - success
 */
bool_t
gen_close_pipe(gen_desc_t *gdp)
{
	cd_state_t	*cdp = gdp->cdp;
	char		*cp;
	int		i,
			ret,
			wstat,
			saverr = 0;
	unsigned int	oa;
	void		(*oh)(int);

	if (!gen_set_eid(cdp))
		return FALSE;

	switch (gdp->fmt) {
#ifdef USE_LIBMP3LAME
	case FILEFMT_MP3:
		(void) gen_halt_lame(gdp);
		break;
#endif	/* USE_LIBMP3LAME */

#ifdef HAS_VORBIS
	case FILEFMT_OGG:
		(void) gen_halt_vorbis(gdp);
		break;
#endif	/* HAS_VORBIS */

	default:
		break;
	}

	/* Close pipe */
	if (gdp->fp != NULL)
		(void) fclose(gdp->fp);
	else
		(void) close(gdp->fd);

	/* Cut off anything beyond argv[0] */
	if ((cp = strchr(gdp->path, ' ')) != NULL)
		*cp = '\0';

	/* Wait for pipe program and get exit status */
	oh = signal(SIGALRM, util_onalrm);
	oa = alarm(1);
	for (i = 0; (ret = WAITPID(gdp->cpid, &wstat, 0)) != gdp->cpid; i++) {
		(void) alarm(0);

		if (ret < 0 && errno != EINTR) {
			saverr = errno;
			break;
		}

		if (i > app_data.hb_timeout) {
			saverr = errno;
			(void) sprintf(cdp->i->msgbuf,
				    "NOTICE: Timed out waiting for program "
				    "\"%s\" to exit (pid=%d).",
				    gdp->path,
				    (int) gdp->cpid
			);
			DBGPRN(DBG_GEN|DBG_SND)(errfp, "%s\n", cdp->i->msgbuf);
			break;
		}

		(void) alarm(1);
	}
	(void) alarm(oa);
	(void) signal(SIGALRM, oh);

	if (ret < 0) {
		if (saverr != EINTR) {
			DBGPRN(DBG_GEN|DBG_SND)(errfp,
				"%s: waitpid failed (pid=%d errno=%d)\n",
				gdp->path, (int) gdp->cpid, saverr);
		}
	}
	else if (WIFEXITED(wstat)) {
		ret = WEXITSTATUS(wstat);
		if (ret == 0) {
			DBGPRN(DBG_GEN|DBG_SND)(errfp,
				"%s: Program exited (pid=%d status=%d)\n",
				gdp->path, (int) gdp->cpid, ret);
		}
		else {
			(void) sprintf(cdp->i->msgbuf,
				"NOTICE: %s\nProgram exited abnormally "
				"(pid=%d status=%d)",
				gdp->path, (int) gdp->cpid, ret
			);
			DBGPRN(DBG_GEN|DBG_SND)(errfp, "%s\n", cdp->i->msgbuf);
		}
	}
	else if (WIFSIGNALED(wstat)) {
		ret = WTERMSIG(wstat);
		(void) sprintf(cdp->i->msgbuf,
			"NOTICE: %s\nProgram killed (pid=%d status=%d)",
			gdp->path, (int) gdp->cpid, ret
		);
		DBGPRN(DBG_GEN|DBG_SND)(errfp, "%s\n", cdp->i->msgbuf);
	}

	MEM_FREE(gdp->path);
	MEM_FREE(gdp);

	(void) gen_reset_eid(cdp);
	return (ret == 0);
}


/*
 * gen_read_chunk
 *	Reads data from the opened file descriptor, catering for possible
 *	interrupts.
 *
 * Args:
 *	gdp	- Pointer to the gen_desc_t structure
 *	data	- Data buffer
 *	len	- Data length
 *
 * Return:
 *	FALSE - failure
 *	TRUE  - success
 */
bool_t
gen_read_chunk(gen_desc_t *gdp, byte_t *data, size_t len)
{
	int	offset = 0,
		cnt = 0,
		n;

	if (data == NULL || len == 0)
		/* Nothing to do */
		return TRUE;

	if (!gen_set_eid(gdp->cdp))
		return FALSE;

	if ((app_data.debug & DBG_SND) != 0) {
		if ((gdp->flags & GDESC_ISPIPE) != 0) {
			char	*cp,
				sav = '\0';

			if ((cp = strchr(gdp->path, ' ')) != NULL) {
				sav = *cp;
				*cp = '\0';
			}

			(void) fprintf(errfp,
					"\nReading pipe [%s]: %d bytes\n",
					gdp->path, (int) len);

			if (cp != NULL)
				*cp = sav;
		}
		else
			(void) fprintf(errfp, "\nReading [%s]: %d bytes\n",
					gdp->path, (int) len);
	}

	/* Read in */
	do {
		errno = 0;
		if (gdp->fp != NULL)
			n = fread(&data[offset], 1, len - offset, gdp->fp);
		else
			n = read(gdp->fd, &data[offset], len - offset);

		/* Have we finished? */
		if (n == (len - offset)) {
			(void) gen_reset_eid(gdp->cdp);
			return TRUE;
		}

		/* Cater for EINTR */
		if (n < 0) {
			if (errno == EINTR) {
				/* Reset and increment counter if not paused */
				n = 0;
				if (gdp->cdp->i->state != CDSTAT_PAUSED)
					cnt++;
			}
			else
				break;
		}

		offset += n;

	} while (offset < len && cnt < CDDA_INTR_MAX);

	(void) sprintf(gdp->cdp->i->msgbuf,
			"gen_read_chunk: read failed (errno=%d)",
			errno);
	DBGPRN(DBG_SND)(errfp, "%s\n", gdp->cdp->i->msgbuf);

	(void) gen_reset_eid(gdp->cdp);
	return FALSE;
}


/*
 * gen_write_chunk
 *	Writes data to the opened file descriptor, catering for possible
 *	interrupts.
 *
 * Args:
 *	gdp	- Pointer to the gen_desc_t structure
 *	data	- Data buffer
 *	len	- Data length
 *
 * Return:
 *	FALSE - failure
 *	TRUE  - success
 */
bool_t
gen_write_chunk(gen_desc_t *gdp, byte_t *data, size_t len)
{
	int	offset = 0,
		cnt = 0,
		n;
	bool_t	ret;

	if (!gen_set_eid(gdp->cdp))
		return FALSE;

	if (data == NULL && len == 0) {
		/* Flush buffered I/O */
		if (gdp->fp != NULL)
			(void) fflush(gdp->fp);

		/* Write EOF */
		ret = !(write(gdp->fd, NULL, 0) < 0);

		(void) gen_reset_eid(gdp->cdp);
		return (ret);
	}

	if (len == 0) {
		/* Nothing to do */
		(void) gen_reset_eid(gdp->cdp);
		return TRUE;
	}

	if ((gdp->flags & GDESC_WRITEOUT) == 0) {
		switch (gdp->fmt) {
#ifdef USE_LIBMP3LAME
		case FILEFMT_MP3:
			ret = gen_encode_lame_chunk(gdp, data, len);
			(void) gen_reset_eid(gdp->cdp);
			return (ret);
#endif	/* USE_LIBMP3LAME */

#ifdef HAS_VORBIS
		case FILEFMT_OGG:
			ret = gen_encode_vorbis_chunk(gdp, data, len);
			(void) gen_reset_eid(gdp->cdp);
			return (ret);
#endif	/* HAS_VORBIS */

		default:
			break;
		}
	}

	gdp->flags &= ~GDESC_WRITEOUT;

	if ((app_data.debug & DBG_SND) != 0) {
		if ((gdp->flags & GDESC_ISPIPE) != 0) {
			char	*cp,
				sav = '\0';

			if ((cp = strchr(gdp->path, ' ')) != NULL) {
				sav = *cp;
				*cp = '\0';
			}

			(void) fprintf(errfp,
					"\nWriting pipe [%s]: %d bytes\n",
					gdp->path, (int) len);

			if (cp != NULL)
				*cp = sav;
		}
		else
			(void) fprintf(errfp, "\nWriting [%s]: %d bytes\n",
					gdp->path, (int) len);
	}

	/* Write out */
	do {
		errno = 0;
		if (gdp->fp != NULL)
			n = fwrite(&data[offset], 1, len - offset, gdp->fp);
		else
			n = write(gdp->fd, &data[offset], len - offset);

		/* Have we finished? */
		if (n == (len - offset)) {
			(void) gen_reset_eid(gdp->cdp);
			return TRUE;
		}

		/* Cater for EINTR */
		if (n < 0) {
			if (errno == EINTR) {
				/* Reset and increment counter if not paused */
				n = 0;
				if (gdp->cdp->i->state != CDSTAT_PAUSED)
					cnt++;
			}
			else
				break;
		}

		offset += n;

	} while (offset < len && cnt < CDDA_INTR_MAX);

	(void) sprintf(gdp->cdp->i->msgbuf,
			"gen_write_chunk: write failed (errno=%d)",
			errno);
	DBGPRN(DBG_SND)(errfp, "%s\n", gdp->cdp->i->msgbuf);

	(void) gen_reset_eid(gdp->cdp);
	return FALSE;
}


/*
 * gen_ioctl
 *	Wrapper for ioctl(2) function.
 *
 * Args:
 *	gdp - Pointer to the gen_desc_t structure
 *	req - ioctl request
 *	arg - ioctl argument
 *
 * Return:
 *	FALSE - failure
 *	TRUE  - success
 */
bool_t
gen_ioctl(gen_desc_t *gdp, int req, void *arg)
{
	/* Flush buffered I/O first */
	if (gdp->fp != NULL)
		(void) fflush(gdp->fp);

	/* Do the ioctl */
	return (!(ioctl(gdp->fd, req, arg) < 0));
}


/*
 * gen_byteswap
 *	Carry out byte swapping.
 *
 * Args:
 *	srcbuf - audio data source buffer
 *	tgtbuf - audio data target buffer
 *	len    - data length in bytes
 *
 * Return:
 *	Nothing.
 */
void
gen_byteswap(byte_t *srcbuf, byte_t *tgtbuf, size_t len)
{
	int	i;

	/* Run through samples */
	for (i = 0; i < (int) len; i += 2) {
		/* Byte swapping */
		tgtbuf[i] = srcbuf[i+1];
		tgtbuf[i+1] = srcbuf[i];;
	}
}


/*
 * gen_chroute
 *	Other optional processing may need be carried out on the audio
 *	prior to sending it to the audio file.
 *
 *	CHROUTE_NORMAL		Leave as is
 *	CHROUTE_REVERSE		Swaps left and right
 *	CHROUTE_L_MONO		Feed left to both channels
 *	CHROUTE_R_MONO	        Feed right to both channels
 *	CHROUTE_MONO		Feed left and right avrg to both channels
 *
 * Args:
 *	data - Data to transform
 *	len  - Data length in bytes
 *
 * Return:
 *	Nothing.
 */
void
gen_chroute(int chroute, sword16_t *data, size_t len)
{
	int		l,
			r;
	sword16_t	temp;

	switch (chroute) {
	case CHROUTE_NORMAL:
		break;

	case CHROUTE_REVERSE:
		for (l = 0; l < len; l += sizeof(sword16_t)) {
			/* Swap left and right samples */
			r = l + 1;
			temp = *(&data[r]);
			*(&data[r]) = *(&data[l]);
			*(&data[l]) = temp;
		}
		break;

	case CHROUTE_L_MONO:
		for (l = 0; l < len; l += sizeof(sword16_t)) {
			/* Make right same as left */
			r = l + 1;
			*(&data[r]) = *(&data[l]);
		}
		break;

	case CHROUTE_R_MONO:
		for (l = 0; l < len; l += sizeof(sword16_t)) {
			/* Make left same as right */
			r = l + 1;
			*(&data[l]) = *(&data[r]);
		}
		break;

	case CHROUTE_MONO:
		for (l = 0; l < len; l += sizeof(sword16_t)) {
			/* Average left and right */
			r = l + 1;
			temp = ((sword16_t) util_lswap16(*(&data[l])) +
				(sword16_t) util_lswap16(*(&data[r]))) >> 1;
			*(&data[l]) = (sword16_t) util_lswap16(temp);
			*(&data[r]) = *(&data[l]);
		}
		break;

	default:
		break;
	}
}


/*
 * gen_write_init
 *	Generic write services initialization
 *
 * Args:
 *	None.
 *
 * Return:
 *	Nothing.
 */
void
gen_write_init(void)
{
	static bool_t	gen_write_initted = FALSE;

	if (!gen_write_initted) {
		gen_write_initted = TRUE;

		ruid = util_get_ouid();
		rgid = util_get_ogid();
		euid = geteuid();
		egid = getegid();
	}
}

#endif	/* CDDA_SYSVIPC */


